"""Cause the Pacemaker Remote connection resource to fail."""

__all__ = ["RemoteRscFailure"]
__copyright__ = "Copyright 2000-2024 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

from pacemaker._cts.tests.remotedriver import RemoteDriver

# Disable various pylint warnings that occur in so many places throughout this
# file it's easiest to just take care of them globally.  This does introduce the
# possibility that we'll miss some other cause of the same warning, but we'll
# just have to be careful.

# pylint doesn't understand that self._env is subscriptable.
# pylint: disable=unsubscriptable-object


class RemoteRscFailure(RemoteDriver):
    """Cause the Pacemaker Remote connection resource to fail."""

    def __init__(self, cm):
        """
        Create a new RemoteRscFailure instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        RemoteDriver.__init__(self, cm)
        self.name = "RemoteRscFailure"

    def __call__(self, node):
        """Perform this test."""
        if not self.start_new_test(node):
            return self.failure(self.fail_string)

        # This is an important step. We are migrating the connection
        # before failing the resource. This verifies that the migration
        # has properly maintained control over the remote-node.
        self.migrate_connection(node)

        self.fail_rsc(node)
        self.cleanup_metal(node)

        self.debug("Waiting for the cluster to recover")
        self._cm.cluster_stable()
        if self.failed:
            return self.failure(self.fail_string)

        return self.success()

    @property
    def errors_to_ignore(self):
        """Return list of errors which should be ignored."""
        return [
            r"schedulerd.*: Recover\s+remote-rsc\s+\(.*\)",
            r"Dummy.*: No process state file found"
        ] + super().errors_to_ignore

    def is_applicable(self):
        """Return True if this test is applicable in the current test configuration."""
        if not RemoteDriver.is_applicable(self):
            return False

        # This test requires at least three nodes: one to convert to a
        # remote node, one to host the connection originally, and one
        # to migrate the connection to.
        return len(self._env["nodes"]) >= 3



"""
A module providing base classes.

These classes are used for defining regression tests and groups of regression
tests.  Everything exported here should be considered an abstract class that
needs to be subclassed in order to do anything useful.  Various functions
will raise NotImplementedError if not overridden by a subclass.
"""

__copyright__ = "Copyright 2009-2024 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+)"

__all__ = ["Test", "Tests"]

import io
import os
import re
import shlex
import signal
import subprocess
import sys
import time

from pacemaker._cts.errors import ExitCodeError, OutputFoundError, OutputNotFoundError, XmlValidationError
from pacemaker._cts.process import pipe_communicate
from pacemaker.buildoptions import BuildOptions
from pacemaker.exitstatus import ExitStatus


def find_validator(rng_file):
    """
    Return the command line used to validate XML output.

    If no validator is found, return None.
    """
    if os.access("/usr/bin/xmllint", os.X_OK):
        if rng_file is None:
            return ["xmllint", "-"]

        return ["xmllint", "--relaxng", rng_file, "-"]

    return None


def rng_directory():
    """Return the directory containing RNG schema files."""
    if "PCMK_schema_directory" in os.environ:
        return os.environ["PCMK_schema_directory"]

    if os.path.exists("%s/cts-fencing.in" % sys.path[0]):
        return "xml"

    return BuildOptions.SCHEMA_DIR


class Pattern:
    """A class for checking log files for a given pattern."""

    def __init__(self, pat, negative=False, regex=False):
        """
        Create a new Pattern instance.

        Arguments:
        pat      -- The string to search for
        negative -- If True, pat must not be found in any input
        regex    -- If True, pat is a regex and not a substring
        """
        self._pat = pat
        self.negative = negative
        self.regex = regex

    def __str__(self):
        return self._pat

    def match(self, line):
        """Return True if this pattern is found in the given line."""
        if self.regex:
            return re.search(self._pat, line) is not None

        return self._pat in line


class Test:
    """
    The base class for a single regression test.

    A single regression test may still run multiple commands as part of its
    execution.
    """

    def __init__(self, name, description, **kwargs):
        """
        Create a new Test instance.

        This method must be provided by all subclasses, which must call
        Test.__init__ first.

        Arguments:
        description -- A user-readable description of the test, helpful in
                       identifying what test is running or has failed.
        name        -- The name of the test.  Command line tools use this
                       attribute to allow running only tests with the exact
                       name, or tests whose name matches a given pattern.
                       This should be unique among all tests.

        Keyword arguments:
        force_wait  --
        logdir      -- The base directory under which to create a directory
                       to store output and temporary data.
        timeout     -- How long to wait for the test to complete.
        verbose     -- Whether to print additional information, including
                       verbose command output and daemon log files.
        """
        self.description = description
        self.executed = False
        self.name = name

        self.force_wait = kwargs.get("force_wait", False)
        self.logdir = kwargs.get("logdir", "/tmp")
        self.timeout = kwargs.get("timeout", 2)
        self.verbose = kwargs.get("verbose", False)

        self._cmds = []
        self._patterns = []

        self._daemon_location = None
        self._daemon_output = ""
        self._daemon_process = None

        self._result_exitcode = ExitStatus.OK
        self._result_txt = ""

    #
    # PROPERTIES
    #

    @property
    def exitcode(self):
        """
        Return the final exitcode of the Test.

        If all commands pass, this property will be ExitStatus.OK.  Otherwise,
        this property will be the exitcode of the first command to fail.
        """
        return self._result_exitcode

    @exitcode.setter
    def exitcode(self, value):
        self._result_exitcode = value

    @property
    def logpath(self):
        """
        Return the path to the log for whatever daemon is being tested.

        Note that this requires all subclasses to set self._daemon_location
        before accessing this property or an exception will be raised.
        """
        return os.path.join(self.logdir, "%s.log" % self._daemon_location)

    #
    # PRIVATE METHODS
    #

    def _kill_daemons(self):
        """Kill any running daemons in preparation for executing the test."""
        raise NotImplementedError("_kill_daemons not provided by subclass")

    def _match_log_patterns(self):
        """
        Check test output for expected patterns.

        Set self.exitcode and self._result_txt as appropriate.  Not all subclass
        will need to do this.
        """
        if len(self._patterns) == 0:
            return

        n_failed_matches = 0
        n_negative_matches = 0

        output = self._daemon_output.split("\n")

        for pat in self._patterns:
            positive_match = False

            for line in output:
                if pat.match(line):
                    if pat.negative:
                        n_negative_matches += 1

                        if self.verbose:
                            print("This pattern should not have matched = '%s" % pat)

                        break

                    positive_match = True
                    break

            if not pat.negative and not positive_match:
                n_failed_matches += 1
                print("Pattern Not Matched = '%s'" % pat)

        if n_failed_matches > 0 or n_negative_matches > 0:
            msg = "FAILURE - '%s' failed. %d patterns out of %d not matched. %d negative matches."
            self._result_txt = msg % (self.name, n_failed_matches, len(self._patterns), n_negative_matches)
            self.exitcode = ExitStatus.ERROR

    def _new_cmd(self, cmd, args, exitcode, **kwargs):
        """
        Add a command to be executed as part of this test.

        Arguments:
        cmd         -- The program to run.
        args        -- Commands line arguments to pass to cmd, as a string.
        exitcode    -- The expected exit code of cmd.  This can be used to
                       run a command that is expected to fail.

        Keyword arguments:
        stdout_match          -- If not None, a string that is expected to be
                                 present in the stdout of cmd.  This can be a
                                 regular expression.
        no_wait               -- Do not wait for cmd to complete.
        stdout_negative_match -- If not None, a string that is expected to be
                                 missing in the stdout of cmd.  This can be a
                                 regualr expression.
        kill                  -- A command to be run after cmd, typically in
                                 order to kill a failed process.  This should be
                                 the entire command line including arguments as
                                 a single string.
        validate              -- If True, the output of cmd will be passed to
                                 xmllint for validation.  If validation fails,
                                 XmlValidationError will be raised.
        check_rng             -- If True and validate is True, command output
                                 will additionally be checked against the
                                 api-result.rng file.
        check_stderr          -- If True, the stderr of cmd will be included in
                                 output.
        env                   -- If not None, variables to set in the environment
        """
        self._cmds.append(
            {
                "args": args,
                "check_rng": kwargs.get("check_rng", True),
                "check_stderr": kwargs.get("check_stderr", True),
                "cmd": cmd,
                "expected_exitcode": exitcode,
                "kill": kwargs.get("kill", None),
                "no_wait": kwargs.get("no_wait", False),
                "stdout_match": kwargs.get("stdout_match", None),
                "stdout_negative_match": kwargs.get("stdout_negative_match", None),
                "validate": kwargs.get("validate", True),
                "env": kwargs.get("env", None),
            }
        )

    def _start_daemons(self):
        """Start any necessary daemons in preparation for executing the test."""
        raise NotImplementedError("_start_daemons not provided by subclass")

    #
    # PUBLIC METHODS
    #

    def add_cmd(self, cmd, args, validate=True, check_rng=True, check_stderr=True,
                env=None):
        """Add a simple command to be executed as part of this test."""
        self._new_cmd(cmd, args, ExitStatus.OK, validate=validate, check_rng=check_rng,
                      check_stderr=check_stderr, env=env)

    def add_cmd_and_kill(self, cmd, args, kill_proc):
        """Add a command and system command to be executed as part of this test."""
        self._new_cmd(cmd, args, ExitStatus.OK, kill=kill_proc)

    def add_cmd_check_stdout(self, cmd, args, match, no_match=None, env=None):
        """Add a simple command with expected output to be executed as part of this test."""
        self._new_cmd(cmd, args, ExitStatus.OK, stdout_match=match,
                      stdout_negative_match=no_match, env=env)

    def add_cmd_expected_fail(self, cmd, args, exitcode=ExitStatus.ERROR):
        """Add a command that is expected to fail to be executed as part of this test."""
        self._new_cmd(cmd, args, exitcode)

    def add_cmd_no_wait(self, cmd, args):
        """Add a simple command to be executed (without waiting) as part of this test."""
        self._new_cmd(cmd, args, ExitStatus.OK, no_wait=True)

    def add_log_pattern(self, pattern, negative=False, regex=False):
        """Add a pattern that should appear in the test's logs."""
        self._patterns.append(Pattern(pattern, negative=negative, regex=regex))

    def _signal_dict(self):
        """Return a dictionary mapping signal numbers to their names."""
        # FIXME: When we support python >= 3.5, this function can be replaced with:
        #   signal.Signals(self.daemon_process.returncode).name
        return {
            getattr(signal, _signame): _signame
            for _signame in dir(signal)
            if _signame.startswith("SIG") and not _signame.startswith("SIG_")
        }

    def clean_environment(self):
        """Clean up the host after executing a test."""
        if self._daemon_process:
            if self._daemon_process.poll() is None:
                self._daemon_process.terminate()
                self._daemon_process.wait()
            else:
                rc = self._daemon_process.returncode
                signame = self._signal_dict().get(-rc, "RET=%s" % rc)
                msg = "FAILURE - '%s' failed. %s abnormally exited during test (%s)."

                self._result_txt = msg % (self.name, self._daemon_location, signame)
                self.exitcode = ExitStatus.ERROR

        self._daemon_process = None
        self._daemon_output = ""

        # the default for utf-8 encoding would error out if e.g. memory corruption
        # makes fenced output any kind of 8 bit value - while still interesting
        # for debugging and we'd still like the regression-test to go over the
        # full set of test-cases
        with open(self.logpath, 'rt', encoding="ISO-8859-1") as logfile:
            for line in logfile.readlines():
                self._daemon_output += line

        if self.verbose:
            print("Daemon Output Start")
            print(self._daemon_output)
            print("Daemon Output End")

    def print_result(self, filler):
        """Print the result of the last test execution."""
        print("%s%s" % (filler, self._result_txt))

    def run(self):
        """Execute this test."""
        i = 1

        self.start_environment()

        if self.verbose:
            print("\n--- START TEST - %s" % self.name)

        self._result_txt = "SUCCESS - '%s'" % (self.name)
        self.exitcode = ExitStatus.OK

        for cmd in self._cmds:
            try:
                self.run_cmd(cmd)
            except ExitCodeError as e:
                print("Step %d FAILED - command returned %s, expected %d" % (i, e, cmd['expected_exitcode']))
                self.set_error(i, cmd)
                break
            except OutputNotFoundError as e:
                print("Step %d FAILED - '%s' was not found in command output: %s" % (i, cmd['stdout_match'], e))
                self.set_error(i, cmd)
                break
            except OutputFoundError as e:
                print("Step %d FAILED - '%s' was found in command output: %s" % (i, cmd['stdout_negative_match'], e))
                self.set_error(i, cmd)
                break
            except XmlValidationError as e:
                print("Step %d FAILED - xmllint failed: %s" % (i, e))
                self.set_error(i, cmd)
                break

            if self.verbose:
                print("Step %d SUCCESS" % (i))

            i += 1

        self.clean_environment()

        if self.exitcode == ExitStatus.OK:
            self._match_log_patterns()

        print(self._result_txt)
        if self.verbose:
            print("--- END TEST - %s\n" % self.name)

        self.executed = True

    def run_cmd(self, args):
        """Execute a command as part of this test."""
        cmd = shlex.split(args['args'])
        cmd.insert(0, args['cmd'])

        if self.verbose:
            print("\n\nRunning: %s" % " ".join(cmd))

        # FIXME: Using "with" here breaks fencing merge tests.
        # pylint: disable=consider-using-with
        if args['env']:
            new_env = os.environ.copy()
            new_env.update(args['env'])
            test = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                    env=new_env)
        else:
            test = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        if args['kill']:
            if self.verbose:
                print("Also running: %s" % args['kill'])

            # Typically, the kill argument is used to detect some sort of
            # failure. Without yielding for a few seconds here, the process
            # launched earlier that is listening for the failure may not have
            # time to connect to pacemaker-execd.
            time.sleep(2)
            subprocess.Popen(shlex.split(args['kill']))

        if not args['no_wait']:
            test.wait()
        else:
            return ExitStatus.OK

        output = pipe_communicate(test, check_stderr=args['check_stderr'])

        if self.verbose:
            print(output)

        if test.returncode != args['expected_exitcode']:
            raise ExitCodeError(test.returncode)

        if args['stdout_match'] is not None and \
           re.search(args['stdout_match'], output) is None:
            raise OutputNotFoundError(output)

        if args['stdout_negative_match'] is not None and \
           re.search(args['stdout_negative_match'], output) is not None:
            raise OutputFoundError(output)

        if args['validate']:
            if args['check_rng']:
                rng_file = "%s/api/api-result.rng" % rng_directory()
            else:
                rng_file = None

            cmd = find_validator(rng_file)
            if not cmd:
                raise XmlValidationError("Could not find validator for %s" % rng_file)

            if self.verbose:
                print("\nRunning: %s" % " ".join(cmd))

            with subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as validator:
                output = pipe_communicate(validator, check_stderr=True, stdin=output)

                if self.verbose:
                    print(output)

                if validator.returncode != 0:
                    raise XmlValidationError(output)

        return ExitStatus.OK

    def set_error(self, step, cmd):
        """Record failure of this test."""
        msg = "FAILURE - '%s' failed at step %d. Command: %s %s"
        self._result_txt = msg % (self.name, step, cmd['cmd'], cmd['args'])
        self.exitcode = ExitStatus.ERROR

    def start_environment(self):
        """Prepare the host for executing a test."""
        if os.path.exists(self.logpath):
            os.remove(self.logpath)

        self._kill_daemons()
        self._start_daemons()

        logfile = None

        init_time = time.time()
        update_time = init_time

        while True:
            # FIXME: Eventually use 'with' here, which seems complicated given
            # everything happens in a loop.
            # pylint: disable=consider-using-with
            time.sleep(0.1)

            if not self.force_wait and logfile is None \
               and os.path.exists(self.logpath):
                logfile = io.open(self.logpath, 'rt', encoding="ISO-8859-1")

            if not self.force_wait and logfile is not None:
                for line in logfile.readlines():
                    if "successfully started" in line:
                        return

            now = time.time()

            if self.timeout > 0 and (now - init_time) >= self.timeout:
                if not self.force_wait:
                    print("\tDaemon %s doesn't seem to have been initialized within %fs."
                          "\n\tConsider specifying a longer '--timeout' value."
                          % (self._daemon_location, self.timeout))
                return

            if self.verbose and (now - update_time) >= 5:
                print("Waiting for %s to be initialized: %fs ..."
                      % (self._daemon_location, now - init_time))
                update_time = now


class Tests:
    """The base class for a collection of regression tests."""

    def __init__(self, **kwargs):
        """
        Create a new Tests instance.

        This method must be provided by all subclasses, which must call
        Tests.__init__ first.

        Keywork arguments:

        force_wait  --
        logdir      -- The base directory under which to create a directory
                       to store output and temporary data.
        timeout     -- How long to wait for the test to complete.
        verbose     -- Whether to print additional information, including
                       verbose command output and daemon log files.
        """
        self.force_wait = kwargs.get("force_wait", False)
        self.logdir = kwargs.get("logdir", "/tmp")
        self.timeout = kwargs.get("timeout", 2)
        self.verbose = kwargs.get("verbose", False)

        self._tests = []

    def exit(self):
        """Exit (with error status code if any test failed)."""
        for test in self._tests:
            if not test.executed:
                continue

            if test.exitcode != ExitStatus.OK:
                sys.exit(ExitStatus.ERROR)

        sys.exit(ExitStatus.OK)

    def print_list(self):
        """List all registered tests."""
        print("\n==== %d TESTS FOUND ====" % len(self._tests))
        print("%35s - %s" % ("TEST NAME", "TEST DESCRIPTION"))
        print("%35s - %s" % ("--------------------", "--------------------"))

        for test in self._tests:
            print("%35s - %s" % (test.name, test.description))

        print("==== END OF LIST ====\n")

    def print_results(self):
        """Print summary of results of executed tests."""
        failures = 0
        success = 0

        print("\n\n======= FINAL RESULTS ==========")
        print("\n--- FAILURE RESULTS:")

        for test in self._tests:
            if not test.executed:
                continue

            if test.exitcode != ExitStatus.OK:
                failures += 1
                test.print_result("    ")
            else:
                success += 1

        if failures == 0:
            print("    None")

        print("\n--- TOTALS\n    Pass:%d\n    Fail:%d\n" % (success, failures))

    def run_single(self, name):
        """Run a single named test."""
        for test in self._tests:
            if test.name == name:
                test.run()
                break

    def run_tests(self):
        """Run all tests."""
        for test in self._tests:
            test.run()

    def run_tests_matching(self, pattern):
        """Run all tests whose name matches a pattern."""
        for test in self._tests:
            if test.name.count(pattern) != 0:
                test.run()



"""Put a node into standby mode and check that resources migrate."""

__all__ = ["StandbyTest"]
__copyright__ = "Copyright 2000-2024 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

from pacemaker._cts.tests.ctstest import CTSTest
from pacemaker._cts.tests.simulstartlite import SimulStartLite
from pacemaker._cts.tests.starttest import StartTest

# Disable various pylint warnings that occur in so many places throughout this
# file it's easiest to just take care of them globally.  This does introduce the
# possibility that we'll miss some other cause of the same warning, but we'll
# just have to be careful.

# pylint doesn't understand that self._env is subscriptable.
# pylint: disable=unsubscriptable-object


class StandbyTest(CTSTest):
    """Put a node into standby and check that resources migrate away from it."""

    def __init__(self, cm):
        """
        Create a new StandbyTest instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        CTSTest.__init__(self, cm)

        self.benchmark = True
        self.name = "Standby"

        self._start = StartTest(cm)
        self._startall = SimulStartLite(cm)

    # make sure the node is active
    # set the node to standby mode
    # check resources, none resource should be running on the node
    # set the node to active mode
    # check resources, resources should have been migrated back (SHOULD THEY?)

    def __call__(self, node):
        """Perform this test."""
        self.incr("calls")
        ret = self._startall(None)
        if not ret:
            return self.failure("Start all nodes failed")

        self.debug("Make sure node %s is active" % node)
        if self._cm.in_standby_mode(node):
            if not self._cm.set_standby_mode(node, False):
                return self.failure("can't set node %s to active mode" % node)

        self._cm.cluster_stable()

        if self._cm.in_standby_mode(node):
            return self.failure("standby status of %s is [on] but we expect [off]" % node)

        watchpats = [
            r"State transition .* -> S_POLICY_ENGINE",
        ]
        watch = self.create_watch(watchpats, self._env["DeadTime"] + 10)
        watch.set_watch()

        self.debug("Setting node %s to standby mode" % node)
        if not self._cm.set_standby_mode(node, True):
            return self.failure("can't set node %s to standby mode" % node)

        self.set_timer("on")

        ret = watch.look_for_all()
        if not ret:
            self._logger.log("Patterns not found: %r" % watch.unmatched)
            self._cm.set_standby_mode(node, False)
            return self.failure("cluster didn't react to standby change on %s" % node)

        self._cm.cluster_stable()

        if not self._cm.in_standby_mode(node):
            return self.failure("standby status of %s is [off] but we expect [on]" % node)

        self.log_timer("on")

        self.debug("Checking resources")
        rscs_on_node = self._cm.active_resources(node)
        if rscs_on_node:
            rc = self.failure("%s set to standby, %r is still running on it" % (node, rscs_on_node))
            self.debug("Setting node %s to active mode" % node)
            self._cm.set_standby_mode(node, False)
            return rc

        self.debug("Setting node %s to active mode" % node)
        if not self._cm.set_standby_mode(node, False):
            return self.failure("can't set node %s to active mode" % node)

        self.set_timer("off")
        self._cm.cluster_stable()

        if self._cm.in_standby_mode(node):
            return self.failure("standby status of %s is [on] but we expect [off]" % node)

        self.log_timer("off")

        return self.success()



"""Stop all running nodes serially."""

__all__ = ["StopOnebyOne"]
__copyright__ = "Copyright 2000-2024 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

from pacemaker._cts.tests.ctstest import CTSTest
from pacemaker._cts.tests.simulstartlite import SimulStartLite
from pacemaker._cts.tests.stoptest import StopTest

# Disable various pylint warnings that occur in so many places throughout this
# file it's easiest to just take care of them globally.  This does introduce the
# possibility that we'll miss some other cause of the same warning, but we'll
# just have to be careful.

# pylint doesn't understand that self._env is subscriptable.
# pylint: disable=unsubscriptable-object


class StopOnebyOne(CTSTest):
    """Stop all running nodes serially."""

    def __init__(self, cm):
        """
        Create a new StartOnebyOne instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        CTSTest.__init__(self, cm)

        self.name = "StopOnebyOne"

        self._startall = SimulStartLite(cm)
        self._stop = StopTest(cm)

    def __call__(self, dummy):
        """Perform this test."""
        self.incr("calls")

        ret = self._startall(None)
        if not ret:
            return self.failure("Setup failed")

        failed = []
        self.set_timer()
        for node in self._env["nodes"]:
            if not self._stop(node):
                failed.append(node)

        if failed:
            return self.failure("Some node failed to stop: %r" % failed)

        return self.success()



"""Test scenario classes for Pacemaker's Cluster Test Suite (CTS)."""

__all__ = [
    "AllOnce",
    "Boot",
    "BootCluster",
    "LeaveBooted",
    "RandomTests",
    "Sequence",
]
__copyright__ = "Copyright 2000-2024 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

import re
import time

from pacemaker._cts.audits import ClusterAudit
from pacemaker._cts.input import should_continue
from pacemaker._cts.tests.ctstest import CTSTest
from pacemaker._cts.watcher import LogWatcher


class ScenarioComponent:
    """
    The base class for all scenario components.

    A scenario component is one single step in a scenario.  Each component is
    basically just a setup and teardown method.
    """

    def __init__(self, cm, env):
        """
        Create a new ScenarioComponent instance.

        Arguments:
        cm  -- A ClusterManager instance
        env -- An Environment instance
        """
        # pylint: disable=invalid-name
        self._cm = cm
        self._env = env

    def is_applicable(self):
        """
        Return True if this component is applicable in the given Environment.

        This method must be provided by all subclasses.
        """
        raise NotImplementedError

    def setup(self):
        """
        Set up the component, returning True on success.

        This method must be provided by all subclasses.
        """
        raise NotImplementedError

    def teardown(self):
        """
        Tear down the given component.

        This method must be provided by all subclasses.
        """
        raise NotImplementedError


class Scenario:
    """
    The base class for scenarios.

    A scenario is an ordered list of ScenarioComponent objects.  A scenario
    proceeds by setting up all its components in sequence, running a list of
    tests and audits, and then tearing down its components in reverse.
    """

    def __init__(self, cm, components, audits, tests):
        """
        Create a new Scenario instance.

        Arguments:
        cm         -- A ClusterManager instance
        components -- A list of ScenarioComponents comprising this Scenario
        audits     -- A list of ClusterAudits that will be performed as
                      part of this Scenario
        tests      -- A list of CTSTests that will be run
        """
        # pylint: disable=invalid-name

        self.stats = {
            "success": 0,
            "failure": 0,
            "BadNews": 0,
            "skipped": 0
        }
        self.tests = tests

        self._audits = audits
        self._bad_news = None
        self._cm = cm
        self._components = components

        for comp in components:
            if not issubclass(comp.__class__, ScenarioComponent):
                raise ValueError("Init value must be subclass of ScenarioComponent")

        for audit in audits:
            if not issubclass(audit.__class__, ClusterAudit):
                raise ValueError("Init value must be subclass of ClusterAudit")

        for test in tests:
            if not issubclass(test.__class__, CTSTest):
                raise ValueError("Init value must be a subclass of CTSTest")

    def is_applicable(self):
        """Return True if all ScenarioComponents are applicable."""
        for comp in self._components:
            if not comp.is_applicable():
                return False

        return True

    def setup(self):
        """
        Set up the scenario, returning True on success.

        If setup fails at some point, tear down those components that did
        successfully set up.
        """
        self._cm.prepare()
        self.audit()  # Also detects remote/local log config
        self._cm.ns.wait_for_all_nodes(self._cm.env["nodes"])

        self.audit()
        self._cm.install_support()

        self._bad_news = LogWatcher(self._cm.env["LogFileName"],
                                    self._cm.templates.get_patterns("BadNews"),
                                    self._cm.env["nodes"],
                                    self._cm.env["log_kind"],
                                    "BadNews", 0)
        self._bad_news.set_watch()  # Call after we've figured out what type of log watching to do in LogAudit

        j = 0
        while j < len(self._components):
            if not self._components[j].setup():
                # OOPS!  We failed.  Tear partial setups down.
                self.audit()
                self._cm.log("Tearing down partial setup")
                self.teardown(j)
                return False

            j += 1

        self.audit()
        return True

    def teardown(self, n_components=None):
        """
        Tear down the scenario in the reverse order it was set up.

        If n_components is not None, only tear down that many components.
        """
        if not n_components:
            n_components = len(self._components) - 1

        j = n_components

        while j >= 0:
            self._components[j].teardown()
            j -= 1

        self.audit()
        self._cm.install_support("uninstall")

    def incr(self, name):
        """Increment the given stats key."""
        if name not in self.stats:
            self.stats[name] = 0

        self.stats[name] += 1

    def run(self, iterations):
        """Run all tests in the scenario the given number of times."""
        self._cm.oprofile_start()

        try:
            self._run_loop(iterations)
            self._cm.oprofile_stop()
        except:  # noqa: E722
            self._cm.oprofile_stop()
            raise

    def _run_loop(self, iterations):
        """Run all the tests the given number of times."""
        raise NotImplementedError

    def run_test(self, test, testcount):
        """
        Run the given test.

        testcount is the number of tests (including this one) that have been
        run across all iterations.
        """
        nodechoice = self._cm.env.random_node()

        ret = True
        did_run = False

        self._cm.clear_instance_errors_to_ignore()
        choice = "(%s)" % nodechoice
        self._cm.log("Running test {:<22} {:<15} [{:>3}]".format(test.name, choice, testcount))

        starttime = test.set_timer()

        if not test.setup(nodechoice):
            self._cm.log("Setup failed")
            ret = False
        else:
            did_run = True
            ret = test(nodechoice)

        if not test.teardown(nodechoice):
            self._cm.log("Teardown failed")

            if not should_continue(self._cm.env):
                raise ValueError("Teardown of %s on %s failed" % (test.name, nodechoice))

            ret = False

        stoptime = time.time()
        self._cm.oprofile_save(testcount)

        elapsed_time = stoptime - starttime
        test_time = stoptime - test.get_timer()

        if "min_time" not in test.stats:
            test.stats["elapsed_time"] = elapsed_time
            test.stats["min_time"] = test_time
            test.stats["max_time"] = test_time
        else:
            test.stats["elapsed_time"] += elapsed_time

            if test_time < test.stats["min_time"]:
                test.stats["min_time"] = test_time

            if test_time > test.stats["max_time"]:
                test.stats["max_time"] = test_time

        if ret:
            self.incr("success")
            test.log_timer()
        else:
            self.incr("failure")
            self._cm.statall()
            did_run = True  # Force the test count to be incremented anyway so test extraction works

        self.audit(test.errors_to_ignore)
        return did_run

    def summarize(self):
        """Output scenario results."""
        self._cm.log("****************")
        self._cm.log("Overall Results:%r" % self.stats)
        self._cm.log("****************")

        stat_filter = {
            "calls": 0,
            "failure": 0,
            "skipped": 0,
            "auditfail": 0,
        }

        self._cm.log("Test Summary")
        for test in self.tests:
            for key in stat_filter:
                stat_filter[key] = test.stats[key]

            name = "Test %s:" % test.name
            self._cm.log("{:<25} {!r}".format(name, stat_filter))

        self._cm.debug("Detailed Results")
        for test in self.tests:
            name = "Test %s:" % test.name
            self._cm.debug("{:<25} {!r}".format(name, stat_filter))

        self._cm.log("<<<<<<<<<<<<<<<< TESTS COMPLETED")

    def audit(self, local_ignore=None):
        """
        Perform all scenario audits and log results.

        If there are too many failures, prompt the user to confirm that the
        scenario should continue running.
        """
        errcount = 0

        ignorelist = ["CTS:"]

        if local_ignore:
            ignorelist.extend(local_ignore)

        ignorelist.extend(self._cm.errors_to_ignore)
        ignorelist.extend(self._cm.instance_errors_to_ignore)

        # This makes sure everything is stabilized before starting...
        failed = 0
        for audit in self._audits:
            if not audit():
                self._cm.log("Audit %s FAILED." % audit.name)
                failed += 1
            else:
                self._cm.debug("Audit %s passed." % audit.name)

        while errcount < 1000:
            match = None
            if self._bad_news:
                match = self._bad_news.look(0)

            if match:
                add_err = True

                for ignore in ignorelist:
                    if add_err and re.search(ignore, match):
                        add_err = False

                if add_err:
                    self._cm.log("BadNews: %s" % match)
                    self.incr("BadNews")
                    errcount += 1
            else:
                break
        else:
            print("Big problems")
            if not should_continue(self._cm.env):
                self._cm.log("Shutting down.")
                self.summarize()
                self.teardown()
                raise ValueError("Looks like we hit a BadNews jackpot!")

        if self._bad_news:
            self._bad_news.end()

        return failed


class AllOnce(Scenario):
    """Every Test Once."""

    def _run_loop(self, iterations):
        testcount = 1

        for test in self.tests:
            self.run_test(test, testcount)
            testcount += 1


class RandomTests(Scenario):
    """Random Test Execution."""

    def _run_loop(self, iterations):
        testcount = 1

        while testcount <= iterations:
            test = self._cm.env.random_gen.choice(self.tests)
            self.run_test(test, testcount)
            testcount += 1


class Sequence(Scenario):
    """Named Tests in Sequence."""

    def _run_loop(self, iterations):
        testcount = 1

        while testcount <= iterations:
            for test in self.tests:
                self.run_test(test, testcount)
                testcount += 1


class Boot(Scenario):
    """Start the Cluster."""

    def _run_loop(self, iterations):
        return


class BootCluster(ScenarioComponent):
    """
    Start the cluster manager on all nodes.

    Wait for each to come up before starting in order to account for the
    possibility that a given node might have been rebooted or crashed
    beforehand.
    """

    def is_applicable(self):
        """Return whether this scenario is applicable."""
        return True

    def setup(self):
        """Set up the component, returning True on success."""
        self._cm.prepare()

        #        Clear out the cobwebs ;-)
        self._cm.stopall(verbose=True, force=True)

        # Now start the Cluster Manager on all the nodes.
        self._cm.log("Starting Cluster Manager on all nodes.")
        return self._cm.startall(verbose=True, quick=True)

    def teardown(self):
        """Tear down the component."""
        self._cm.log("Stopping Cluster Manager on all nodes")
        self._cm.stopall(verbose=True, force=False)


class LeaveBooted(BootCluster):
    """Leave all nodes up when the scenario is complete."""

    def teardown(self):
        """Tear down the component."""
        self._cm.log("Leaving Cluster running on all nodes")



"""Stop and restart a node."""

__all__ = ["RestartTest"]
__copyright__ = "Copyright 2000-2024 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

from pacemaker._cts.tests.ctstest import CTSTest
from pacemaker._cts.tests.starttest import StartTest
from pacemaker._cts.tests.stoptest import StopTest


class RestartTest(CTSTest):
    """Stop and restart a node."""

    def __init__(self, cm):
        """
        Create a new RestartTest instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        CTSTest.__init__(self, cm)
        self.benchmark = True
        self.name = "Restart"

        self._start = StartTest(cm)
        self._stop = StopTest(cm)

    def __call__(self, node):
        """Perform this test."""
        self.incr("calls")
        self.incr("node:%s" % node)

        if self._cm.stat_cm(node):
            self.incr("WasStopped")
            if not self._start(node):
                return self.failure("start (setup) failure: %s" % node)

        self.set_timer()

        if not self._stop(node):
            return self.failure("stop failure: %s" % node)

        if not self._start(node):
            return self.failure("start failure: %s" % node)

        return self.success()



"""Remote command runner for Pacemaker's Cluster Test Suite (CTS)."""

__all__ = ["RemoteExec", "RemoteFactory"]
__copyright__ = "Copyright 2014-2024 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

import re
import os

from subprocess import Popen, PIPE
from threading import Thread

from pacemaker._cts.logging import LogFactory


def convert2string(lines):
    """
    Convert byte strings to UTF-8 strings.

    Lists of byte strings are converted to a list of UTF-8 strings.  All other
    text formats are passed through.
    """
    if isinstance(lines, bytes):
        return lines.decode("utf-8")

    if isinstance(lines, list):
        lst = []
        for line in lines:
            if isinstance(line, bytes):
                line = line.decode("utf-8")

            lst.append(line)

        return lst

    return lines


class AsyncCmd(Thread):
    """A class for doing the hard work of running a command on another machine."""

    def __init__(self, node, command, proc=None, delegate=None):
        """
        Create a new AsyncCmd instance.

        Arguments:
        node     -- The remote machine to run on
        command  -- The ssh command string to use for remote execution
        proc     -- If not None, a process object previously created with Popen.
                    Instead of spawning a new process, we will then wait on
                    this process to finish and handle its output.
        delegate -- When the command completes, call the async_complete method
                    on this object
        """
        self._command = command
        self._delegate = delegate
        self._logger = LogFactory()
        self._node = node
        self._proc = proc

        Thread.__init__(self)

    def run(self):
        """Run the previously instantiated AsyncCmd object."""
        out = None
        err = None

        if not self._proc:
            # pylint: disable=consider-using-with
            self._proc = Popen(self._command, stdout=PIPE, stderr=PIPE, close_fds=True, shell=True)

        self._logger.debug("cmd: async: target=%s, pid=%d: %s" % (self._node, self._proc.pid, self._command))
        self._proc.wait()

        if self._delegate:
            self._logger.debug("cmd: pid %d returned %d to %r" % (self._proc.pid, self._proc.returncode, self._delegate))
        else:
            self._logger.debug("cmd: pid %d returned %d" % (self._proc.pid, self._proc.returncode))

        if self._proc.stderr:
            err = self._proc.stderr.readlines()
            self._proc.stderr.close()

            for line in err:
                self._logger.debug("cmd: stderr[%d]: %s" % (self._proc.pid, line))

            err = convert2string(err)

        if self._proc.stdout:
            out = self._proc.stdout.readlines()
            self._proc.stdout.close()
            out = convert2string(out)

        if self._delegate:
            self._delegate.async_complete(self._proc.pid, self._proc.returncode, out, err)


class RemoteExec:
    """
    An abstract class for remote execution.

    It runs a command on another machine using ssh and scp.
    """

    def __init__(self, command, cp_command, silent=False):
        """
        Create a new RemoteExec instance.

        Arguments:
        command    -- The ssh command string to use for remote execution
        cp_command -- The scp command string to use for copying files
        silent     -- Should we log command status?
        """
        self._command = command
        self._cp_command = cp_command
        self._logger = LogFactory()
        self._silent = silent
        self._our_node = os.uname()[1].lower()

    def _fixcmd(self, cmd):
        """Perform shell escapes on certain characters in the input cmd string."""
        return re.sub("\'", "'\\''", cmd)

    def _cmd(self, args):
        """Given a list of arguments, return the string that will be run on the remote system."""
        sysname = args[0]
        command = args[1]

        if sysname is None or sysname.lower() in [self._our_node, "localhost"]:
            ret = command
        else:
            ret = "%s %s '%s'" % (self._command, sysname, self._fixcmd(command))

        return ret

    def _log(self, args):
        """Log a message."""
        if not self._silent:
            self._logger.log(args)

    def _debug(self, args):
        """Log a message at the debug level."""
        if not self._silent:
            self._logger.debug(args)

    def call_async(self, node, command, delegate=None):
        """
        Run the given command on the given remote system and do not wait for it to complete.

        Arguments:
        node     -- The remote machine to run on
        command  -- The command to run, as a string
        delegate -- When the command completes, call the async_complete method
                    on this object

        Returns the running process object.
        """
        aproc = AsyncCmd(node, self._cmd([node, command]), delegate=delegate)
        aproc.start()
        return aproc

    def __call__(self, node, command, synchronous=True, verbose=2):
        """
        Run the given command on the given remote system.

        If you call this class like a function, this is what gets called.  It's
        approximately the same as a system() call on the remote machine.

        Arguments:
        node        -- The remote machine to run on
        command     -- The command to run, as a string
        synchronous -- Should we wait for the command to complete?
        verbose     -- If 0, do not lo:g anything.  If 1, log the command and its
                       return code but not its output.  If 2, additionally log
                       command output.

        Returns a tuple of (return code, command output).
        """
        rc = 0
        result = None
        # pylint: disable=consider-using-with
        proc = Popen(self._cmd([node, command]),
                     stdout=PIPE, stderr=PIPE, close_fds=True, shell=True)

        if not synchronous and proc.pid > 0 and not self._silent:
            aproc = AsyncCmd(node, command, proc=proc)
            aproc.start()
            return (rc, result)

        if proc.stdout:
            result = proc.stdout.readlines()
            proc.stdout.close()
        else:
            self._log("No stdout stream")

        rc = proc.wait()

        if verbose > 0:
            self._debug("cmd: target=%s, rc=%d: %s" % (node, rc, command))

        result = convert2string(result)

        if proc.stderr:
            errors = proc.stderr.readlines()
            proc.stderr.close()

            for err in errors:
                self._debug("cmd: stderr: %s" % err)

        if verbose == 2:
            for line in result:
                self._debug("cmd: stdout: %s" % line)

        return (rc, result)

    def copy(self, source, target, silent=False):
        """
        Perform a copy of the source file to the remote target.

        This function uses the cp_command provided when the RemoteExec object
        was created.

        Returns the return code of the cp_command.
        """
        cmd = "%s '%s' '%s'" % (self._cp_command, source, target)
        rc = os.system(cmd)

        if not silent:
            self._debug("cmd: rc=%d: %s" % (rc, cmd))

        return rc

    def exists_on_all(self, filename, hosts):
        """Return True if specified file exists on all specified hosts."""
        for host in hosts:
            rc = self(host, "test -r %s" % filename)
            if rc != 0:
                return False

        return True


class RemoteFactory:
    """A class for constructing a singleton instance of a RemoteExec object."""

    # Class variables

    # -n: no stdin, -x: no X11,
    # -o ServerAliveInterval=5: disconnect after 3*5s if the server
    # stops responding
    command = ("ssh -l root -n -x -o ServerAliveInterval=5 "
               "-o ConnectTimeout=10 -o TCPKeepAlive=yes "
               "-o ServerAliveCountMax=3 ")

    # -B: batch mode, -q: no stats (quiet)
    cp_command = "scp -B -q"

    instance = None

    # pylint: disable=invalid-name
    def getInstance(self):
        """
        Return the previously created instance of RemoteExec.

        If no instance exists, create one and then return that.
        """
        if not RemoteFactory.instance:
            RemoteFactory.instance = RemoteExec(RemoteFactory.command,
                                                RemoteFactory.cp_command,
                                                False)
        return RemoteFactory.instance

    def enable_qarsh(self):
        """Enable the QA remote shell."""
        # http://nstraz.wordpress.com/2008/12/03/introducing-qarsh/
        print("Using QARSH for connections to cluster nodes")

        RemoteFactory.command = "qarsh -t 300 -l root"
        RemoteFactory.cp_command = "qacp -q"



"""Stop the cluster manager on a given node."""

__all__ = ["StopTest"]
__copyright__ = "Copyright 2000-2024 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

from pacemaker._cts.tests.ctstest import CTSTest

# Disable various pylint warnings that occur in so many places throughout this
# file it's easiest to just take care of them globally.  This does introduce the
# possibility that we'll miss some other cause of the same warning, but we'll
# just have to be careful.

# pylint doesn't understand that self._rsh is callable.
# pylint: disable=not-callable
# pylint doesn't understand that self._env is subscriptable.
# pylint: disable=unsubscriptable-object


class StopTest(CTSTest):
    """
    A pseudo-test that sets up conditions before running some other test.

    This class stops the cluster manager on a given node.  Other test classes
    should not use this one as a superclass.
    """

    def __init__(self, cm):
        """
        Create a new StopTest instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        CTSTest.__init__(self, cm)
        self.name = "Stop"

    def __call__(self, node):
        """Stop the given node, returning whether this succeeded or not."""
        self.incr("calls")
        if self._cm.expected_status[node] != "up":
            return self.skipped()

        # Technically we should always be able to notice ourselves stopping
        patterns = [
            self.templates["Pat:We_stopped"] % node,
        ]

        # Any active node needs to notice this one left
        # (note that this won't work if we have multiple partitions)
        for other in self._env["nodes"]:
            if self._cm.expected_status[other] == "up" and other != node:
                patterns.append(self.templates["Pat:They_stopped"] % (other, node))

        watch = self.create_watch(patterns, self._env["DeadTime"])
        watch.set_watch()

        if node == self._cm.our_node:
            self.incr("us")
        else:
            if self._cm.upcount() <= 1:
                self.incr("all")
            else:
                self.incr("them")

        self._cm.stop_cm(node)
        watch.look_for_all()

        failreason = None
        unmatched_str = "||"

        if watch.unmatched:
            (_, output) = self._rsh(node, "/bin/ps axf", verbose=1)
            for line in output:
                self.debug(line)

            (_, output) = self._rsh(node, "/usr/sbin/dlm_tool dump 2>/dev/null", verbose=1)
            for line in output:
                self.debug(line)

            for regex in watch.unmatched:
                self._logger.log("ERROR: Shutdown pattern not found: %s" % regex)
                unmatched_str += "%s||" % regex
                failreason = "Missing shutdown pattern"

        self._cm.cluster_stable(self._env["DeadTime"])

        if not watch.unmatched or self._cm.upcount() == 0:
            return self.success()

        if len(watch.unmatched) >= self._cm.upcount():
            return self.failure("no match against (%s)" % unmatched_str)

        if failreason is None:
            return self.success()

        return self.failure(failreason)



# These warnings are not useful in unit tests.
# pylint: disable=missing-class-docstring,missing-function-docstring,missing-module-docstring

__copyright__ = "Copyright 2023 the Pacemaker project contributors"
__license__ = "GPLv2+"

import unittest

from pacemaker._cts.network import next_ip


# next_ip makes a bunch of assumptions that we are not going to test here:
#
# * The env argument actually contains an "IPBase" key with a string in it
# * The string is a properly formatted IPv4 or IPv6 address, with no extra
#   leading or trailing whitespace

class NextIPTestCase(unittest.TestCase):
    def test_ipv4(self):
        # The first time next_ip is called, it will read the IPBase out of
        # the environment.  After that, it just goes off whatever it
        # previously returned, so the environment value doesn't matter.
        self.assertEqual(next_ip("1.1.1.1"), "1.1.1.2")
        self.assertEqual(next_ip(), "1.1.1.3")

        # Passing reset=True will force it to re-read the environment.  Here,
        # we use that to ask it for a value outside of the available range.
        self.assertRaises(ValueError, next_ip, "1.1.1.255", reset=True)

    def test_ipv6(self):
        # Same comments as for the test_ipv4 function, plus we need to reset
        # here because otherwise it will remember what happened in that function.
        self.assertEqual(next_ip("fe80::fc54:ff:fe09:101", reset=True),
                         "fe80::fc54:ff:fe09:102")
        self.assertEqual(next_ip(),
                         "fe80::fc54:ff:fe09:103")

        self.assertRaises(ValueError, next_ip, "fe80::fc54:ff:fe09:ffff", reset=True)



"""Network related utilities for CTS."""

__all__ = ["next_ip"]
__copyright__ = "Copyright 2000-2024 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

# pylint: disable=global-statement
CURRENT_IP = None


def next_ip(ip_base=None, reset=False):
    """
    Return the next available IP address.

    This function only increments the last portion of the IP address.  Once it
    has hit the upper limit, ValueError will be raised.

    Arguments:
    ip_base -- The initial IP address to start from.  The first call to next_ip
               will return the next IP address from this base.  Each subsequent
               call will return the next address from the previous call, so you
               can just omit this argument for subsequent calls.
    reset   -- Force next_ip to start from ip_base again.  This requires also
               passing the ip_base argument.  (Mostly useful for unit testing,
               but may be useful elsewhere).
    """
    global CURRENT_IP

    if CURRENT_IP is None or reset:
        CURRENT_IP = ip_base

    new_ip = None

    # Split the existing IP address up into a tuple of:
    # (everything except the last part of the addr, the separator, the last part of the addr).
    # For instance, "192.168.1.2" becomes ("192.168.1", ".", "2").  Then,
    # increment the last part of the address and paste everything back
    # together.
    if ":" in CURRENT_IP:
        # This is an IPv6 address
        fields = CURRENT_IP.rpartition(":")
        new_ip = int(fields[2], 16) + 1

        if new_ip > 0xffff:
            raise ValueError("No more available IP addresses")

        # hex() puts "0x" at the front of the string, so strip it off.
        new_ip = hex(new_ip)[2:]

    else:
        # This is an IPv4 address
        fields = CURRENT_IP.rpartition(".")
        new_ip = int(fields[2]) + 1

        if new_ip > 255:
            raise ValueError("No more available IP addresses")

    CURRENT_IP = "%s%s%s" % (fields[0], fields[1], new_ip)
    return CURRENT_IP



"""Main classes for Pacemaker's Cluster Test Suite (CTS)."""

__all__ = ["CtsLab", "NodeStatus", "Process"]
__copyright__ = "Copyright 2000-2024 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

import sys
import time
import traceback

from pacemaker.exitstatus import ExitStatus
from pacemaker._cts.environment import EnvFactory
from pacemaker._cts.input import should_continue
from pacemaker._cts.logging import LogFactory
from pacemaker._cts.remote import RemoteFactory


class CtsLab:
    """
    A class that defines the Lab Environment for the Cluster Test System.

    It defines those things which are expected to change from test
    environment to test environment for the same cluster manager.

    This is where you define the set of nodes that are in your test lab,
    what kind of reset mechanism you use, etc.  All this data is stored
    as key/value pairs in an Environment instance constructed from arguments
    passed to this class.

    The CTS code ignores names it doesn't know about or need.  Individual
    tests have access to this information, and it is perfectly acceptable
    to provide hints, tweaks, fine-tuning directions, or other information
    to the tests through this mechanism.
    """

    def __init__(self, args=None):
        """
        Create a new CtsLab instance.

        This class can be treated kind of like a dictionary due to the presence
        of typical dict functions like __contains__, __getitem__, and __setitem__.
        However, it is not a dictionary so do not rely on standard dictionary
        behavior.

        Arguments:
        args -- A list of command line parameters, minus the program name.
        """
        self._env = EnvFactory().getInstance(args)
        self._logger = LogFactory()

    def dump(self):
        """Print the current environment."""
        self._env.dump()

    def __contains__(self, key):
        """Return True if the given environment key exists."""
        # pylint gets confused because of EnvFactory here.
        # pylint: disable=unsupported-membership-test
        return key in self._env

    def __getitem__(self, key):
        """Return the given environment key, or raise KeyError if it does not exist."""
        # Throughout this file, pylint has trouble understanding that EnvFactory
        # and RemoteFactory are singleton instances that can be treated as callable
        # and subscriptable objects.  Various warnings are disabled because of this.
        # See also a comment about self._rsh in environment.py.
        # pylint: disable=unsubscriptable-object
        return self._env[key]

    def __setitem__(self, key, value):
        """Set the given environment key to the given value, overriding any previous value."""
        # pylint: disable=unsupported-assignment-operation
        self._env[key] = value

    def run(self, scenario, iterations):
        """
        Run the given scenario the given number of times.

        Returns ExitStatus.OK on success, or ExitStatus.ERROR on error.
        """
        if not scenario:
            self._logger.log("No scenario was defined")
            return ExitStatus.ERROR

        self._logger.log("Cluster nodes: ")
        # pylint: disable=unsubscriptable-object
        for node in self._env["nodes"]:
            self._logger.log("    * %s" % (node))

        if not scenario.setup():
            return ExitStatus.ERROR

        # We want to alert on any exceptions caused by running a scenario, so
        # here it's okay to disable the pylint warning.
        # pylint: disable=bare-except
        try:
            scenario.run(iterations)
        except:  # noqa: E722
            self._logger.log("Exception by %s" % sys.exc_info()[0])
            self._logger.traceback(traceback)

            scenario.summarize()
            scenario.teardown()
            return ExitStatus.ERROR

        scenario.teardown()
        scenario.summarize()

        if scenario.stats["failure"] > 0:
            return ExitStatus.ERROR

        if scenario.stats["success"] != iterations:
            self._logger.log("No failure count but success != requested iterations")
            return ExitStatus.ERROR

        return ExitStatus.OK


class NodeStatus:
    """
    A class for querying the status of cluster nodes.

    Are nodes up?  Do they respond to SSH connections?
    """

    def __init__(self, env):
        """
        Create a new NodeStatus instance.

        Arguments:
        env -- An Environment instance
        """
        self._env = env

    def _node_booted(self, node):
        """Return True if the given node is booted (responds to pings)."""
        # pylint: disable=not-callable
        (rc, _) = RemoteFactory().getInstance()("localhost", "ping -nq -c1 -w1 %s" % node, verbose=0)
        return rc == 0

    def _sshd_up(self, node):
        """Return true if sshd responds on the given node."""
        # pylint: disable=not-callable
        (rc, _) = RemoteFactory().getInstance()(node, "true", verbose=0)
        return rc == 0

    def wait_for_node(self, node, timeout=300):
        """
        Wait for a node to become available.

        Should the timeout be reached, the user will be given a choice whether
        to continue or not.  If not, ValueError will be raised.

        Returns True when the node is available, or False if the timeout is
        reached.
        """
        initial_timeout = timeout
        anytimeouts = False

        while timeout > 0:
            if self._node_booted(node) and self._sshd_up(node):
                if anytimeouts:
                    # Fudge to wait for the system to finish coming up
                    time.sleep(30)
                    LogFactory().debug("Node %s now up" % node)

                return True

            time.sleep(30)
            if not anytimeouts:
                LogFactory().debug("Waiting for node %s to come up" % node)

            anytimeouts = True
            timeout -= 1

        LogFactory().log("%s did not come up within %d tries" % (node, initial_timeout))
        if not should_continue(self._env["continue"]):
            raise ValueError("%s did not come up within %d tries" % (node, initial_timeout))

        return False

    def wait_for_all_nodes(self, nodes, timeout=300):
        """Return True when all nodes come up, or False if the timeout is reached."""
        for node in nodes:
            if not self.wait_for_node(node, timeout):
                return False

        return True


class Process:
    """A class for managing a Pacemaker daemon."""

    # pylint: disable=invalid-name
    def __init__(self, cm, name, dc_only=False, pats=None, dc_pats=None,
                 badnews_ignore=None):
        """
        Create a new Process instance.

        Arguments:
        cm              -- A ClusterManager instance
        name            -- The command being run
        dc_only         -- Should this daemon be killed only on the DC?
        pats            -- Regexes we expect to find in log files
        dc_pats         -- Additional DC-specific regexes we expect to find
                           in log files
        badnews_ignore  -- Regexes for lines in the log that can be ignored
        """
        self._cm = cm
        self.badnews_ignore = badnews_ignore
        self.dc_only = dc_only
        self.dc_pats = dc_pats
        self.name = name
        self.pats = pats

        if self.badnews_ignore is None:
            self.badnews_ignore = []

        if self.dc_pats is None:
            self.dc_pats = []

        if self.pats is None:
            self.pats = []

    def kill(self, node):
        """Kill the instance of this process running on the given node."""
        (rc, _) = self._cm.rsh(node, "killall -9 %s" % self.name)

        if rc != 0:
            self._cm.log("ERROR: Kill %s failed on node %s" % (self.name, node))



"""Fail the connection resource and fence the remote node."""

__all__ = ["RemoteStonithd"]
__copyright__ = "Copyright 2000-2024 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

from pacemaker._cts.tests.remotedriver import RemoteDriver


class RemoteStonithd(RemoteDriver):
    """Fail the connection resource and fence the remote node."""

    def __init__(self, cm):
        """
        Create a new RemoteStonithd instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        RemoteDriver.__init__(self, cm)

        self.name = "RemoteStonithd"

    def __call__(self, node):
        """Perform this test."""
        if not self.start_new_test(node):
            return self.failure(self.fail_string)

        self.fail_connection(node)
        self.cleanup_metal(node)

        self.debug("Waiting for the cluster to recover")
        self._cm.cluster_stable()
        if self.failed:
            return self.failure(self.fail_string)

        return self.success()

    def is_applicable(self):
        """Return True if this test is applicable in the current test configuration."""
        if not RemoteDriver.is_applicable(self):
            return False

        return self._env.get("DoFencing", True)

    @property
    def errors_to_ignore(self):
        """Return list of errors which should be ignored."""
        return [
            r"Lost connection to Pacemaker Remote node",
            r"Software caused connection abort",
            r"pacemaker-controld.*:\s+error.*: Operation remote-.*_monitor",
            r"pacemaker-controld.*:\s+error.*: Result of monitor operation for remote-.*",
            r"schedulerd.*:\s+Recover\s+remote-.*\s+\(.*\)",
            r"error: Result of monitor operation for .* on remote-.*: Internal communication failure"
        ] + super().errors_to_ignore



"""A module providing custom exception classes used throughout the pacemaker library."""

__all__ = ["ExitCodeError", "OutputFoundError", "OutputNotFoundError", "XmlValidationError"]
__copyright__ = "Copyright 2009-2024 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+)"


class TestError(Exception):
    """Base class for exceptions in this module."""


class ExitCodeError(TestError):
    """Exception raised when command exit status is unexpected."""

    def __init__(self, exit_code):
        """Create a new ExitCodeError instance."""
        TestError.__init__(self)
        self.exit_code = exit_code

    def __str__(self):
        """Return a printable string for this exception."""
        return repr(self.exit_code)


class OutputNotFoundError(TestError):
    """Exception raised when command output does not contain wanted string."""

    def __init__(self, output):
        """Create a new OutputNotFoundError instance."""
        TestError.__init__(self)
        self.output = output

    def __str__(self):
        """Return a printable string for this exception."""
        return repr(self.output)


class OutputFoundError(TestError):
    """Exception raised when command output contains unwanted string."""

    def __init__(self, output):
        """Create a new OutputFoundError instance."""
        TestError.__init__(self)
        self.output = output

    def __str__(self):
        """Return a printable string for this exception."""
        return repr(self.output)


class XmlValidationError(TestError):
    """Exception raised when xmllint fails."""

    def __init__(self, output):
        """Create a new XmlValidationError instance."""
        TestError.__init__(self)
        self.output = output

    def __str__(self):
        """Return a printable string for this exception."""
        return repr(self.output)



"""Simultaneously stop running nodes."""

__all__ = ["SimulStopLite"]
__copyright__ = "Copyright 2000-2024 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

from pacemaker._cts.tests.ctstest import CTSTest

# Disable various pylint warnings that occur in so many places throughout this
# file it's easiest to just take care of them globally.  This does introduce the
# possibility that we'll miss some other cause of the same warning, but we'll
# just have to be careful.

# pylint doesn't understand that self._rsh is callable.
# pylint: disable=not-callable
# pylint doesn't understand that self._env is subscriptable.
# pylint: disable=unsubscriptable-object


class SimulStopLite(CTSTest):
    """
    A pseudo-test that sets up conditions before running some other test.

    This class stops any running nodes more or less simultaneously.  It can be
    used both to set up a test or to clean up a test.  Other test classes
    should not use this one as a superclass.
    """

    def __init__(self, cm):
        """
        Create a new SimulStopLite instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        CTSTest.__init__(self, cm)
        self.name = "SimulStopLite"

    def __call__(self, dummy):
        """Return whether stopping all running nodes more or less simultaneously succeeds."""
        self.incr("calls")
        self.debug("Setup: %s" % self.name)

        # We ignore the "node" parameter...
        watchpats = []

        for node in self._env["nodes"]:
            if self._cm.expected_status[node] == "up":
                self.incr("WasStarted")
                watchpats.append(self.templates["Pat:We_stopped"] % node)

        if len(watchpats) == 0:
            return self.success()

        # Stop all the nodes - at about the same time...
        watch = self.create_watch(watchpats, self._env["DeadTime"] + 10)

        watch.set_watch()
        self.set_timer()
        for node in self._env["nodes"]:
            if self._cm.expected_status[node] == "up":
                self._cm.stop_cm_async(node)

        if watch.look_for_all():
            # Make sure they're completely down with no residule
            for node in self._env["nodes"]:
                self._rsh(node, self.templates["StopCmd"])

            return self.success()

        did_fail = False
        up_nodes = []
        for node in self._env["nodes"]:
            if self._cm.stat_cm(node):
                did_fail = True
                up_nodes.append(node)

        if did_fail:
            return self.failure("Active nodes exist: %s" % up_nodes)

        self._logger.log("Warn: All nodes stopped but CTS didn't detect: %s" % watch.unmatched)
        return self.failure("Missing log message: %s " % watch.unmatched)

    def is_applicable(self):
        """Return True if this test is applicable in the current test configuration."""
        return False



"""Create a split brain cluster and verify a resource is multiply managed."""

__all__ = ["SplitBrainTest"]
__copyright__ = "Copyright 2000-2024 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

import time

from pacemaker._cts.input import should_continue
from pacemaker._cts.tests.ctstest import CTSTest
from pacemaker._cts.tests.simulstartlite import SimulStartLite
from pacemaker._cts.tests.starttest import StartTest

# Disable various pylint warnings that occur in so many places throughout this
# file it's easiest to just take care of them globally.  This does introduce the
# possibility that we'll miss some other cause of the same warning, but we'll
# just have to be careful.

# pylint doesn't understand that self._rsh is callable.
# pylint: disable=not-callable
# pylint doesn't understand that self._env is subscriptable.
# pylint: disable=unsubscriptable-object


class SplitBrainTest(CTSTest):
    """
    Create a split brain cluster.

    This test verifies that one node in each partition takes over the
    resource, resulting in two nodes running the same resource.
    """

    def __init__(self, cm):
        """
        Create a new SplitBrainTest instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        CTSTest.__init__(self, cm)

        self.is_experimental = True
        self.name = "SplitBrain"

        self._start = StartTest(cm)
        self._startall = SimulStartLite(cm)

    def _isolate_partition(self, partition):
        """Create a new partition containing the given nodes."""
        other_nodes = self._env["nodes"].copy()

        for node in partition:
            try:
                other_nodes.remove(node)
            except ValueError:
                self._logger.log("Node %s not in %r from %r" % (node, self._env["nodes"], partition))

        if not other_nodes:
            return

        self.debug("Creating partition: %r" % partition)
        self.debug("Everyone else: %r" % other_nodes)

        for node in partition:
            if not self._cm.isolate_node(node, other_nodes):
                self._logger.log("Could not isolate %s" % node)
                return

    def _heal_partition(self, partition):
        """Move the given nodes out of their own partition back into the cluster."""
        other_nodes = self._env["nodes"].copy()

        for node in partition:
            try:
                other_nodes.remove(node)
            except ValueError:
                self._logger.log("Node %s not in %r" % (node, self._env["nodes"]))

        if len(other_nodes) == 0:
            return

        self.debug("Healing partition: %r" % partition)
        self.debug("Everyone else: %r" % other_nodes)

        for node in partition:
            self._cm.unisolate_node(node, other_nodes)

    def __call__(self, node):
        """Perform this test."""
        self.incr("calls")
        self.passed = True
        partitions = {}

        if not self._startall(None):
            return self.failure("Setup failed")

        while True:
            # Retry until we get multiple partitions
            partitions = {}
            p_max = len(self._env["nodes"])

            for n in self._env["nodes"]:
                p = self._env.random_gen.randint(1, p_max)

                if p not in partitions:
                    partitions[p] = []

                partitions[p].append(n)

            p_max = len(partitions)
            if p_max > 1:
                break
            # else, try again

        self.debug("Created %d partitions" % p_max)
        for (key, val) in partitions.items():
            self.debug("Partition[%s]:\t%r" % (key, val))

        # Disabling STONITH to reduce test complexity for now
        self._rsh(node, "crm_attribute -V -n stonith-enabled -v false")

        for val in partitions.values():
            self._isolate_partition(val)

        count = 30
        while count > 0:
            if len(self._cm.find_partitions()) != p_max:
                time.sleep(10)
            else:
                break
        else:
            self.failure("Expected partitions were not created")

        # Target number of partitions formed - wait for stability
        if not self._cm.cluster_stable():
            self.failure("Partitioned cluster not stable")

        # Now audit the cluster state
        self._cm.partitions_expected = p_max
        if not self.audit():
            self.failure("Audits failed")

        self._cm.partitions_expected = 1

        # And heal them again
        for val in partitions.values():
            self._heal_partition(val)

        # Wait for a single partition to form
        count = 30
        while count > 0:
            if len(self._cm.find_partitions()) != 1:
                time.sleep(10)
                count -= 1
            else:
                break
        else:
            self.failure("Cluster did not reform")

        # Wait for it to have the right number of members
        count = 30
        while count > 0:
            members = []

            partitions = self._cm.find_partitions()
            if partitions:
                members = partitions[0].split()

            if len(members) != len(self._env["nodes"]):
                time.sleep(10)
                count -= 1
            else:
                break
        else:
            self.failure("Cluster did not completely reform")

        # Wait up to 20 minutes - the delay is more preferable than
        # trying to continue with in a messed up state
        if not self._cm.cluster_stable(1200):
            self.failure("Reformed cluster not stable")

            if not should_continue(self._env):
                raise ValueError("Reformed cluster not stable")

        # Turn fencing back on
        if self._env["DoFencing"]:
            self._rsh(node, "crm_attribute -V -D -n stonith-enabled")

        self._cm.cluster_stable()

        if self.passed:
            return self.success()

        return self.failure("See previous errors")

    @property
    def errors_to_ignore(self):
        """Return a list of errors which should be ignored."""
        return [
            r"Another DC detected:",
            r"(ERROR|error).*: .*Application of an update diff failed",
            r"pacemaker-controld.*:.*not in our membership list",
            r"CRIT:.*node.*returning after partition"
        ]

    def is_applicable(self):
        """Return True if this test is applicable in the current test configuration."""
        if not CTSTest.is_applicable(self):
            return False

        return len(self._env["nodes"]) > 2



# These warnings are not useful in unit tests.
# pylint: disable=missing-class-docstring,missing-function-docstring,missing-module-docstring

__copyright__ = "Copyright 2023 the Pacemaker project contributors"
__license__ = "GPLv2+"

import unittest

from pacemaker.exitstatus import ExitStatus


class ExitStatusTestCase(unittest.TestCase):
    def test_min_max(self):
        self.assertEqual(ExitStatus.OK, 0)
        self.assertEqual(ExitStatus.MAX, 255)



"""Base classes for CTS tests."""

__all__ = ["CTSTest"]
__copyright__ = "Copyright 2000-2024 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

import re

from pacemaker._cts.environment import EnvFactory
from pacemaker._cts.logging import LogFactory
from pacemaker._cts.patterns import PatternSelector
from pacemaker._cts.remote import RemoteFactory
from pacemaker._cts.timer import Timer
from pacemaker._cts.watcher import LogWatcher

# Disable various pylint warnings that occur in so many places throughout this
# file it's easiest to just take care of them globally.  This does introduce the
# possibility that we'll miss some other cause of the same warning, but we'll
# just have to be careful.


class CTSTest:
    """
    The base class for all cluster tests.

    This implements a basic set of properties and behaviors like setup, tear
    down, time keeping, and statistics tracking.  It is up to specific tests
    to implement their own specialized behavior on top of this class.
    """

    def __init__(self, cm):
        """
        Create a new CTSTest instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        # pylint: disable=invalid-name

        self.audits = []
        self.name = None
        self.templates = PatternSelector(cm["Name"])

        self.stats = {
            "auditfail": 0,
            "calls": 0,
            "failure": 0,
            "skipped": 0,
            "success": 0
        }

        self._cm = cm
        self._env = EnvFactory().getInstance()
        self._r_o2cb = None
        self._r_ocfs2 = []
        self._rsh = RemoteFactory().getInstance()
        self._logger = LogFactory()
        self._timers = {}

        self.benchmark = True  # which tests to benchmark
        self.failed = False
        self.is_experimental = False
        self.is_loop = False
        self.is_unsafe = False
        self.is_valgrind = False
        self.passed = True

    def log(self, args):
        """Log a message."""
        self._logger.log(args)

    def debug(self, args):
        """Log a debug message."""
        self._logger.debug(args)

    def get_timer(self, key="test"):
        """Get the start time of the given timer."""
        try:
            return self._timers[key].start_time
        except KeyError:
            return 0

    def set_timer(self, key="test"):
        """Set the start time of the given timer to now, and return that time."""
        if key not in self._timers:
            self._timers[key] = Timer(self._logger, self.name, key)

        self._timers[key].start()
        return self._timers[key].start_time

    def log_timer(self, key="test"):
        """Log the elapsed time of the given timer."""
        if key not in self._timers:
            return

        elapsed = self._timers[key].elapsed
        self.debug("%s:%s runtime: %.2f" % (self.name, key, elapsed))
        del self._timers[key]

    def incr(self, name):
        """Increment the given stats key."""
        if name not in self.stats:
            self.stats[name] = 0

        self.stats[name] += 1

        # Reset the test passed boolean
        if name == "calls":
            self.passed = True

    def failure(self, reason="none"):
        """Increment the failure count, with an optional failure reason."""
        self.passed = False
        self.incr("failure")
        self._logger.log(("Test %s" % self.name).ljust(35) + " FAILED: %s" % reason)

        return False

    def success(self):
        """Increment the success count."""
        self.incr("success")
        return True

    def skipped(self):
        """Increment the skipped count."""
        self.incr("skipped")
        return True

    def __call__(self, node):
        """Perform this test."""
        raise NotImplementedError

    def audit(self):
        """Perform all the relevant audits (see ClusterAudit), returning whether or not they all passed."""
        passed = True

        for audit in self.audits:
            if not audit():
                self._logger.log("Internal %s Audit %s FAILED." % (self.name, audit.name))
                self.incr("auditfail")
                passed = False

        return passed

    def setup(self, node):
        """Set up this test."""
        # node is used in subclasses
        # pylint: disable=unused-argument

        return self.success()

    def teardown(self, node):
        """Tear down this test."""
        # node is used in subclasses
        # pylint: disable=unused-argument

        return self.success()

    def create_watch(self, patterns, timeout, name=None):
        """
        Create a new LogWatcher object.

        This object can be used to search log files for matching patterns
        during this test's run.

        Arguments:
        patterns -- A list of regular expressions to match against the log
        timeout  -- Default number of seconds to watch a log file at a time;
                    this can be overridden by the timeout= parameter to
                    self.look on an as-needed basis
        name     -- A unique name to use when logging about this watch
        """
        if not name:
            name = self.name

        return LogWatcher(self._env["LogFileName"], patterns,
                          self._env["nodes"], self._env["log_kind"], name,
                          timeout)

    def local_badnews(self, prefix, watch, local_ignore=None):
        """
        Search through log files for messages.

        Arguments:
        prefix       -- The string to look for at the beginning of lines,
                        or "LocalBadNews:" if None.
        watch        -- The LogWatcher object to use for searching.
        local_ignore -- A list of regexes that, if found in a line, will
                        cause that line to be ignored.

        Return the number of matches found.
        """
        errcount = 0
        if not prefix:
            prefix = "LocalBadNews:"

        ignorelist = [" CTS: ", prefix]

        if local_ignore:
            ignorelist += local_ignore

        while errcount < 100:
            match = watch.look(0)
            if match:
                add_err = True

                for ignore in ignorelist:
                    if add_err and re.search(ignore, match):
                        add_err = False

                if add_err:
                    self._logger.log("%s %s" % (prefix, match))
                    errcount += 1
            else:
                break
        else:
            self._logger.log("Too many errors!")

        watch.end()
        return errcount

    def is_applicable(self):
        """
        Return True if this test is applicable in the current test configuration.

        This method must be implemented by all subclasses.
        """
        if self.is_loop and not self._env["loop-tests"]:
            return False

        if self.is_unsafe and not self._env["unsafe-tests"]:
            return False

        if self.is_valgrind and not self._env["valgrind-tests"]:
            return False

        if self.is_experimental and not self._env["experimental-tests"]:
            return False

        if self._env["benchmark"] and not self.benchmark:
            return False

        return True

    @property
    def errors_to_ignore(self):
        """Return a list of errors which should be ignored."""
        return []



"""ClusterManager class for Pacemaker's Cluster Test Suite (CTS)."""

__all__ = ["ClusterManager"]
__copyright__ = """Copyright 2000-2024 the Pacemaker project contributors.
Certain portions by Huang Zhen <zhenhltc@cn.ibm.com> are copyright 2004
International Business Machines. The version control history for this file
may have further details."""
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

import os
import re
import time

from collections import UserDict

from pacemaker.buildoptions import BuildOptions
from pacemaker._cts.CTS import NodeStatus
from pacemaker._cts.audits import AuditResource
from pacemaker._cts.cib import ConfigFactory
from pacemaker._cts.environment import EnvFactory
from pacemaker._cts.logging import LogFactory
from pacemaker._cts.patterns import PatternSelector
from pacemaker._cts.remote import RemoteFactory
from pacemaker._cts.watcher import LogWatcher

# pylint doesn't understand that self._rsh is callable (it stores the
# singleton instance of RemoteExec, as returned by the getInstance method
# of RemoteFactory).  It's possible we could fix this with type annotations,
# but those were introduced with python 3.5 and we only support python 3.4.
# I think we could also fix this by getting rid of the getInstance methods,
# but that's a project for another day.  For now, just disable the warning.
# pylint: disable=not-callable

# ClusterManager has a lot of methods.
# pylint: disable=too-many-public-methods


class ClusterManager(UserDict):
    """
    An abstract base class for managing the cluster.

    This class implements high-level operations on the cluster and/or its cluster
    managers.  Actual cluster-specific management classes should be subclassed
    from this one.

    Among other things, this class tracks the state every node is expected to be in.
    """

    def _final_conditions(self):
        """Check all keys to make sure they have a non-None value."""
        for (key, val) in self._data.items():
            if val is None:
                raise ValueError("Improper derivation: self[%s] must be overridden by subclass." % key)

    def __init__(self):
        """
        Create a new ClusterManager instance.

        This class can be treated kind of like a dictionary due to the process
        of certain dict functions like __getitem__ and __setitem__.  This is
        because it contains a lot of name/value pairs.  However, it is not
        actually a dictionary so do not rely on standard dictionary behavior.
        """
        # Eventually, ClusterManager should not be a UserDict subclass.  Until
        # that point...
        # pylint: disable=super-init-not-called
        self.__instance_errors_to_ignore = []

        self._cib_installed = False
        self._data = {}
        self._logger = LogFactory()

        self.env = EnvFactory().getInstance()
        self.expected_status = {}
        self.name = self.env["Name"]
        # pylint: disable=invalid-name
        self.ns = NodeStatus(self.env)
        self.our_node = os.uname()[1].lower()
        self.partitions_expected = 1
        self.rsh = RemoteFactory().getInstance()
        self.templates = PatternSelector(self.env["Name"])

        self._final_conditions()

        self._cib_factory = ConfigFactory(self)
        self._cib = self._cib_factory.create_config(self.env["Schema"])
        self._cib_sync = {}

    def __getitem__(self, key):
        """
        Return the given key, checking for it in several places.

        If key is "Name", return the name of the cluster manager.  If the key
        was previously added to the dictionary via __setitem__, return that.
        Otherwise, return the template pattern for the key.

        This method should not be used and may be removed in the future.
        """
        if key == "Name":
            return self.name

        print("FIXME: Getting %s from %r" % (key, self))
        if key in self._data:
            return self._data[key]

        return self.templates.get_patterns(key)

    def __setitem__(self, key, value):
        """
        Set the given key to the given value, overriding any previous value.

        This method should not be used and may be removed in the future.
        """
        print("FIXME: Setting %s=%s on %r" % (key, value, self))
        self._data[key] = value

    def clear_instance_errors_to_ignore(self):
        """Reset instance-specific errors to ignore on each iteration."""
        self.__instance_errors_to_ignore = []

    @property
    def instance_errors_to_ignore(self):
        """Return a list of known errors that should be ignored for a specific test instance."""
        return self.__instance_errors_to_ignore

    @property
    def errors_to_ignore(self):
        """Return a list of known error messages that should be ignored."""
        return self.templates.get_patterns("BadNewsIgnore")

    def log(self, args):
        """Log a message."""
        self._logger.log(args)

    def debug(self, args):
        """Log a debug message."""
        self._logger.debug(args)

    def upcount(self):
        """Return how many nodes are up."""
        count = 0

        for node in self.env["nodes"]:
            if self.expected_status[node] == "up":
                count += 1

        return count

    def install_support(self, command="install"):
        """
        Install or uninstall the CTS support files.

        This includes various init scripts and data, daemons, fencing agents, etc.
        """
        for node in self.env["nodes"]:
            self.rsh(node, "%s/cts-support %s" % (BuildOptions.DAEMON_DIR, command))

    def prepare_fencing_watcher(self):
        """Return a LogWatcher object that watches for fencing log messages."""
        # If we don't have quorum now but get it as a result of starting this node,
        # then a bunch of nodes might get fenced
        if self.has_quorum(None):
            self.debug("Have quorum")
            return None

        if not self.templates["Pat:Fencing_start"]:
            print("No start pattern")
            return None

        if not self.templates["Pat:Fencing_ok"]:
            print("No ok pattern")
            return None

        stonith = None
        stonith_pats = []
        for peer in self.env["nodes"]:
            if self.expected_status[peer] == "up":
                continue

            stonith_pats.extend([
                self.templates["Pat:Fencing_ok"] % peer,
                self.templates["Pat:Fencing_start"] % peer,
            ])

        stonith = LogWatcher(self.env["LogFileName"], stonith_pats, self.env["nodes"],
                             self.env["log_kind"], "StartupFencing", 0)
        stonith.set_watch()
        return stonith

    def fencing_cleanup(self, node, stonith):
        """Wait for a previously fenced node to return to the cluster."""
        peer_list = []
        peer_state = {}

        self.debug("Looking for nodes that were fenced as a result of %s starting" % node)

        # If we just started a node, we may now have quorum (and permission to fence)
        if not stonith:
            self.debug("Nothing to do")
            return peer_list

        q = self.has_quorum(None)
        if not q and len(self.env["nodes"]) > 2:
            # We didn't gain quorum - we shouldn't have shot anyone
            self.debug("Quorum: %s Len: %d" % (q, len(self.env["nodes"])))
            return peer_list

        for n in self.env["nodes"]:
            peer_state[n] = "unknown"

        # Now see if any states need to be updated
        self.debug("looking for: %r" % stonith.regexes)
        shot = stonith.look(0)

        while shot:
            self.debug("Found: %r" % shot)
            del stonith.regexes[stonith.whichmatch]

            # Extract node name
            for n in self.env["nodes"]:
                if re.search(self.templates["Pat:Fencing_ok"] % n, shot):
                    peer = n
                    peer_state[peer] = "complete"
                    self.__instance_errors_to_ignore.append(self.templates["Pat:Fencing_ok"] % peer)

                elif peer_state[n] != "complete" and re.search(self.templates["Pat:Fencing_start"] % n, shot):
                    # TODO: Correctly detect multiple fencing operations for the same host
                    peer = n
                    peer_state[peer] = "in-progress"
                    self.__instance_errors_to_ignore.append(self.templates["Pat:Fencing_start"] % peer)

            if not peer:
                self._logger.log("ERROR: Unknown stonith match: %r" % shot)

            elif peer not in peer_list:
                self.debug("Found peer: %s" % peer)
                peer_list.append(peer)

            # Get the next one
            shot = stonith.look(60)

        for peer in peer_list:
            self.debug("   Peer %s was fenced as a result of %s starting: %s" % (peer, node, peer_state[peer]))
            if self.env["at-boot"]:
                self.expected_status[peer] = "up"
            else:
                self.expected_status[peer] = "down"

            if peer_state[peer] == "in-progress":
                # Wait for any in-progress operations to complete
                shot = stonith.look(60)

                while stonith.regexes and shot:
                    self.debug("Found: %r" % shot)
                    del stonith.regexes[stonith.whichmatch]
                    shot = stonith.look(60)

            # Now make sure the node is alive too
            self.ns.wait_for_node(peer, self.env["DeadTime"])

            # Poll until it comes up
            if self.env["at-boot"]:
                if not self.stat_cm(peer):
                    time.sleep(self.env["StartTime"])

                if not self.stat_cm(peer):
                    self._logger.log("ERROR: Peer %s failed to restart after being fenced" % peer)
                    return None

        return peer_list

    def start_cm(self, node, verbose=False):
        """Start up the cluster manager on a given node."""
        if verbose:
            self._logger.log("Starting %s on node %s" % (self.templates["Name"], node))
        else:
            self.debug("Starting %s on node %s" % (self.templates["Name"], node))

        if node not in self.expected_status:
            self.expected_status[node] = "down"

        if self.expected_status[node] != "down":
            return True

        # Technically we should always be able to notice ourselves starting
        patterns = [
            self.templates["Pat:Local_started"] % node,
        ]

        if self.upcount() == 0:
            patterns.append(self.templates["Pat:DC_started"] % node)
        else:
            patterns.append(self.templates["Pat:NonDC_started"] % node)

        watch = LogWatcher(self.env["LogFileName"], patterns,
                           self.env["nodes"], self.env["log_kind"],
                           "StartaCM", self.env["StartTime"] + 10)

        self.install_config(node)

        self.expected_status[node] = "any"

        if self.stat_cm(node) and self.cluster_stable(self.env["DeadTime"]):
            self._logger.log("%s was already started" % node)
            return True

        stonith = self.prepare_fencing_watcher()
        watch.set_watch()

        (rc, _) = self.rsh(node, self.templates["StartCmd"])
        if rc != 0:
            self._logger.log("Warn: Start command failed on node %s" % node)
            self.fencing_cleanup(node, stonith)
            return False

        self.expected_status[node] = "up"
        watch_result = watch.look_for_all()

        if watch.unmatched:
            for regex in watch.unmatched:
                self._logger.log("Warn: Startup pattern not found: %s" % regex)

        if watch_result and self.cluster_stable(self.env["DeadTime"]):
            self.fencing_cleanup(node, stonith)
            return True

        if self.stat_cm(node) and self.cluster_stable(self.env["DeadTime"]):
            self.fencing_cleanup(node, stonith)
            return True

        self._logger.log("Warn: Start failed for node %s" % node)
        return False

    def start_cm_async(self, node, verbose=False):
        """Start up the cluster manager on a given node without blocking."""
        if verbose:
            self._logger.log("Starting %s on node %s" % (self["Name"], node))
        else:
            self.debug("Starting %s on node %s" % (self["Name"], node))

        self.install_config(node)
        self.rsh(node, self.templates["StartCmd"], synchronous=False)
        self.expected_status[node] = "up"

    def stop_cm(self, node, verbose=False, force=False):
        """Stop the cluster manager on a given node."""
        if verbose:
            self._logger.log("Stopping %s on node %s" % (self["Name"], node))
        else:
            self.debug("Stopping %s on node %s" % (self["Name"], node))

        if self.expected_status[node] != "up" and not force:
            return True

        (rc, _) = self.rsh(node, self.templates["StopCmd"])
        if rc == 0:
            # Make sure we can continue even if corosync leaks
            self.expected_status[node] = "down"
            self.cluster_stable(self.env["DeadTime"])
            return True

        self._logger.log("ERROR: Could not stop %s on node %s" % (self["Name"], node))
        return False

    def stop_cm_async(self, node):
        """Stop the cluster manager on a given node without blocking."""
        self.debug("Stopping %s on node %s" % (self["Name"], node))

        self.rsh(node, self.templates["StopCmd"], synchronous=False)
        self.expected_status[node] = "down"

    def startall(self, nodelist=None, verbose=False, quick=False):
        """Start the cluster manager on every node in the cluster, or on every node in nodelist."""
        if not nodelist:
            nodelist = self.env["nodes"]

        for node in nodelist:
            if self.expected_status[node] == "down":
                self.ns.wait_for_all_nodes(nodelist, 300)

        if not quick:
            # This is used for "basic sanity checks", so only start one node ...
            return self.start_cm(nodelist[0], verbose=verbose)

        # Approximation of SimulStartList for --boot
        watchpats = [
            self.templates["Pat:DC_IDLE"],
        ]
        for node in nodelist:
            watchpats.extend([
                self.templates["Pat:InfraUp"] % node,
                self.templates["Pat:PacemakerUp"] % node,
                self.templates["Pat:Local_started"] % node,
                self.templates["Pat:They_up"] % (nodelist[0], node),
            ])

        #   Start all the nodes - at about the same time...
        watch = LogWatcher(self.env["LogFileName"], watchpats, self.env["nodes"],
                           self.env["log_kind"], "fast-start",
                           self.env["DeadTime"] + 10)
        watch.set_watch()

        if not self.start_cm(nodelist[0], verbose=verbose):
            return False

        for node in nodelist:
            self.start_cm_async(node, verbose=verbose)

        watch.look_for_all()
        if watch.unmatched:
            for regex in watch.unmatched:
                self._logger.log("Warn: Startup pattern not found: %s" % regex)

        if not self.cluster_stable():
            self._logger.log("Cluster did not stabilize")
            return False

        return True

    def stopall(self, nodelist=None, verbose=False, force=False):
        """Stop the cluster manager on every node in the cluster, or on every node in nodelist."""
        ret = True

        if not nodelist:
            nodelist = self.env["nodes"]

        for node in self.env["nodes"]:
            if self.expected_status[node] == "up" or force:
                if not self.stop_cm(node, verbose=verbose, force=force):
                    ret = False

        return ret

    def statall(self, nodelist=None):
        """Return the status of the cluster manager on every node in the cluster, or on every node in nodelist."""
        result = {}

        if not nodelist:
            nodelist = self.env["nodes"]

        for node in nodelist:
            if self.stat_cm(node):
                result[node] = "up"
            else:
                result[node] = "down"

        return result

    def isolate_node(self, target, nodes=None):
        """Break communication between the target node and all other nodes in the cluster, or nodes."""
        if not nodes:
            nodes = self.env["nodes"]

        for node in nodes:
            if node == target:
                continue

            (rc, _) = self.rsh(target, self.templates["BreakCommCmd"] % node)
            if rc != 0:
                self._logger.log("Could not break the communication between %s and %s: %d" % (target, node, rc))
                return False

            self.debug("Communication cut between %s and %s" % (target, node))

        return True

    def unisolate_node(self, target, nodes=None):
        """Re-establish communication between the target node and all other nodes in the cluster, or nodes."""
        if not nodes:
            nodes = self.env["nodes"]

        for node in nodes:
            if node == target:
                continue

            # Limit the amount of time we have asynchronous connectivity for
            # Restore both sides as simultaneously as possible
            self.rsh(target, self.templates["FixCommCmd"] % node, synchronous=False)
            self.rsh(node, self.templates["FixCommCmd"] % target, synchronous=False)
            self.debug("Communication restored between %s and %s" % (target, node))

    def oprofile_start(self, node=None):
        """Start profiling on the given node, or all nodes in the cluster."""
        if not node:
            for n in self.env["oprofile"]:
                self.oprofile_start(n)

        elif node in self.env["oprofile"]:
            self.debug("Enabling oprofile on %s" % node)
            self.rsh(node, "opcontrol --init")
            self.rsh(node, "opcontrol --setup --no-vmlinux --separate=lib --callgraph=20 --image=all")
            self.rsh(node, "opcontrol --start")
            self.rsh(node, "opcontrol --reset")

    def oprofile_save(self, test, node=None):
        """Save profiling data and restart profiling on the given node, or all nodes in the cluster."""
        if not node:
            for n in self.env["oprofile"]:
                self.oprofile_save(test, n)

        elif node in self.env["oprofile"]:
            self.rsh(node, "opcontrol --dump")
            self.rsh(node, "opcontrol --save=cts.%d" % test)
            # Read back with: opreport -l session:cts.0 image:<directory>/c*
            self.oprofile_stop(node)
            self.oprofile_start(node)

    def oprofile_stop(self, node=None):
        """
        Start profiling on the given node, or all nodes in the cluster.

        This does not save profiling data, so call oprofile_save first if needed.
        """
        if not node:
            for n in self.env["oprofile"]:
                self.oprofile_stop(n)

        elif node in self.env["oprofile"]:
            self.debug("Stopping oprofile on %s" % node)
            self.rsh(node, "opcontrol --reset")
            self.rsh(node, "opcontrol --shutdown 2>&1 > /dev/null")

    def install_config(self, node):
        """Remove and re-install the CIB on the first node in the cluster."""
        if not self.ns.wait_for_node(node):
            self.log("Node %s is not up." % node)
            return

        if node in self._cib_sync or not self.env["ClobberCIB"]:
            return

        self._cib_sync[node] = True
        self.rsh(node, "rm -f %s/cib*" % BuildOptions.CIB_DIR)

        # Only install the CIB on the first node, all the other ones will pick it up from there
        if self._cib_installed:
            return

        self._cib_installed = True
        if self.env["CIBfilename"] is None:
            self.log("Installing Generated CIB on node %s" % node)
            self._cib.install(node)

        else:
            self.log("Installing CIB (%s) on node %s" % (self.env["CIBfilename"], node))

            rc = self.rsh.copy(self.env["CIBfilename"], "root@" + (self.templates["CIBfile"] % node))

            if rc != 0:
                raise ValueError("Can not scp file to %s %d" % (node, rc))

        self.rsh(node, "chown %s %s/cib.xml" % (BuildOptions.DAEMON_USER, BuildOptions.CIB_DIR))

    def prepare(self):
        """
        Finish initialization.

        Clear out the expected status and record the current status of every
        node in the cluster.
        """
        self.partitions_expected = 1
        for node in self.env["nodes"]:
            self.expected_status[node] = ""

            if self.env["experimental-tests"]:
                self.unisolate_node(node)

            self.stat_cm(node)

    def test_node_cm(self, node):
        """
        Check the status of a given node.

        Returns 0 if the node is down, 1 if the node is up but unstable, and 2
        if the node is up and stable.
        """
        watchpats = [
            "Current ping state: (S_IDLE|S_NOT_DC)",
            self.templates["Pat:NonDC_started"] % node,
            self.templates["Pat:DC_started"] % node,
        ]

        idle_watch = LogWatcher(self.env["LogFileName"], watchpats, [node],
                                self.env["log_kind"], "ClusterIdle")
        idle_watch.set_watch()

        (_, out) = self.rsh(node, self.templates["StatusCmd"] % node, verbose=1)

        if not out:
            out = ""
        else:
            out = out[0].strip()

        self.debug("Node %s status: '%s'" % (node, out))

        if out.find('ok') < 0:
            if self.expected_status[node] == "up":
                self.log("Node status for %s is %s but we think it should be %s"
                         % (node, "down", self.expected_status[node]))

            self.expected_status[node] = "down"
            return 0

        if self.expected_status[node] == "down":
            self.log("Node status for %s is %s but we think it should be %s: %s"
                     % (node, "up", self.expected_status[node], out))

        self.expected_status[node] = "up"

        # check the output first - because syslog-ng loses messages
        if out.find('S_NOT_DC') != -1:
            # Up and stable
            return 2

        if out.find('S_IDLE') != -1:
            # Up and stable
            return 2

        # fall back to syslog-ng and wait
        if not idle_watch.look():
            # just up
            self.debug("Warn: Node %s is unstable: %s" % (node, out))
            return 1

        # Up and stable
        return 2

    def stat_cm(self, node):
        """Report the status of the cluster manager on a given node."""
        return self.test_node_cm(node) > 0

    # Being up and being stable is not the same question...
    def node_stable(self, node):
        """Return whether or not the given node is stable."""
        if self.test_node_cm(node) == 2:
            return True

        self.log("Warn: Node %s not stable" % node)
        return False

    def partition_stable(self, nodes, timeout=None):
        """Return whether or not all nodes in the given partition are stable."""
        watchpats = [
            "Current ping state: S_IDLE",
            self.templates["Pat:DC_IDLE"],
        ]

        self.debug("Waiting for cluster stability...")

        if timeout is None:
            timeout = self.env["DeadTime"]

        if len(nodes) < 3:
            self.debug("Cluster is inactive")
            return True

        idle_watch = LogWatcher(self.env["LogFileName"], watchpats, nodes.split(),
                                self.env["log_kind"], "ClusterStable", timeout)
        idle_watch.set_watch()

        for node in nodes.split():
            # have each node dump its current state
            self.rsh(node, self.templates["StatusCmd"] % node, verbose=1)

        ret = idle_watch.look()

        while ret:
            self.debug(ret)

            for node in nodes.split():
                if re.search(node, ret):
                    return True

            ret = idle_watch.look()

        self.debug("Warn: Partition %r not IDLE after %ds" % (nodes, timeout))
        return False

    def cluster_stable(self, timeout=None, double_check=False):
        """Return whether or not all nodes in the cluster are stable."""
        partitions = self.find_partitions()

        for partition in partitions:
            if not self.partition_stable(partition, timeout):
                return False

        if not double_check:
            return True

        # Make sure we are really stable and that all resources,
        # including those that depend on transient node attributes,
        # are started if they were going to be
        time.sleep(5)
        for partition in partitions:
            if not self.partition_stable(partition, timeout):
                return False

        return True

    def is_node_dc(self, node, status_line=None):
        """
        Return whether or not the given node is the cluster DC.

        Check the given status_line, or query the cluster if None.
        """
        if not status_line:
            (_, out) = self.rsh(node, self.templates["StatusCmd"] % node, verbose=1)

            if out:
                status_line = out[0].strip()

        if not status_line:
            return False

        if status_line.find('S_IDLE') != -1:
            return True

        if status_line.find('S_INTEGRATION') != -1:
            return True

        if status_line.find('S_FINALIZE_JOIN') != -1:
            return True

        if status_line.find('S_POLICY_ENGINE') != -1:
            return True

        if status_line.find('S_TRANSITION_ENGINE') != -1:
            return True

        return False

    def active_resources(self, node):
        """Return a list of primitive resources active on the given node."""
        (_, output) = self.rsh(node, "crm_resource -c", verbose=1)
        resources = []
        for line in output:
            if not re.search("^Resource", line):
                continue

            tmp = AuditResource(self, line)
            if tmp.type == "primitive" and tmp.host == node:
                resources.append(tmp.id)

        return resources

    def resource_location(self, rid):
        """Return a list of nodes on which the given resource is running."""
        resource_nodes = []
        for node in self.env["nodes"]:
            if self.expected_status[node] != "up":
                continue

            cmd = self.templates["RscRunning"] % rid
            (rc, lines) = self.rsh(node, cmd)

            if rc == 127:
                self.log("Command '%s' failed. Binary or pacemaker-cts package not installed?" % cmd)
                for line in lines:
                    self.log("Output: %s " % line)

            elif rc == 0:
                resource_nodes.append(node)

        return resource_nodes

    def find_partitions(self):
        """
        Return a list of all partitions in the cluster.

        Each element of the list is itself a list of all active nodes in that
        partition.
        """
        ccm_partitions = []

        for node in self.env["nodes"]:
            if self.expected_status[node] != "up":
                self.debug("Node %s is down... skipping" % node)
                continue

            (_, out) = self.rsh(node, self.templates["PartitionCmd"], verbose=1)

            if not out:
                self.log("no partition details for %s" % node)
                continue

            partition = out[0].strip()

            if len(partition) <= 2:
                self.log("bad partition details for %s" % node)
                continue

            nodes = partition.split()
            nodes.sort()
            partition = ' '.join(nodes)

            found = 0
            for a_partition in ccm_partitions:
                if partition == a_partition:
                    found = 1

            if found == 0:
                self.debug("Adding partition from %s: %s" % (node, partition))
                ccm_partitions.append(partition)
            else:
                self.debug("Partition '%s' from %s is consistent with existing entries" % (partition, node))

        self.debug("Found partitions: %r" % ccm_partitions)
        return ccm_partitions

    def has_quorum(self, node_list):
        """Return whether or not the cluster has quorum."""
        # If we are auditing a partition, then one side will
        #   have quorum and the other not.
        # So the caller needs to tell us which we are checking
        # If no value for node_list is specified... assume all nodes
        if not node_list:
            node_list = self.env["nodes"]

        for node in node_list:
            if self.expected_status[node] != "up":
                continue

            (_, quorum) = self.rsh(node, self.templates["QuorumCmd"], verbose=1)
            quorum = quorum[0].strip()

            if quorum.find("1") != -1:
                return True

            if quorum.find("0") != -1:
                return False

            self.debug("WARN: Unexpected quorum test result from %s:%s" % (node, quorum))

        return False

    @property
    def components(self):
        """
        Return a list of all patterns that should be ignored for the cluster's components.

        This must be provided by all subclasses.
        """
        raise NotImplementedError

    def in_standby_mode(self, node):
        """Return whether or not the node is in Standby."""
        (_, out) = self.rsh(node, self.templates["StandbyQueryCmd"] % node, verbose=1)

        if not out:
            return False

        out = out[0].strip()
        self.debug("Standby result: %s" % out)
        return out == "on"

    def set_standby_mode(self, node, status):
        """
        Set node to Standby if status is True, or Active if status is False.

        Return whether the node is now in the requested status.
        """
        current_status = self.in_standby_mode(node)

        if current_status == status:
            return True

        if status:
            cmd = self.templates["StandbyCmd"] % (node, "on")
        else:
            cmd = self.templates["StandbyCmd"] % (node, "off")

        (rc, _) = self.rsh(node, cmd)
        return rc == 0

    def add_dummy_rsc(self, node, rid):
        """Add a dummy resource with the given ID to the given node."""
        rsc_xml = """ '<resources>
                <primitive class=\"ocf\" id=\"%s\" provider=\"pacemaker\" type=\"Dummy\">
                    <operations>
                        <op id=\"%s-interval-10s\" interval=\"10s\" name=\"monitor\"/
                    </operations>
                </primitive>
            </resources>'""" % (rid, rid)
        constraint_xml = """ '<constraints>
                <rsc_location id=\"location-%s-%s\" node=\"%s\" rsc=\"%s\" score=\"INFINITY\"/>
            </constraints>'
            """ % (rid, node, node, rid)

        self.rsh(node, self.templates['CibAddXml'] % rsc_xml)
        self.rsh(node, self.templates['CibAddXml'] % constraint_xml)

    def remove_dummy_rsc(self, node, rid):
        """Remove the previously added dummy resource given by rid on the given node."""
        constraint = "\"//rsc_location[@rsc='%s']\"" % rid
        rsc = "\"//primitive[@id='%s']\"" % rid

        self.rsh(node, self.templates['CibDelXpath'] % constraint)
        self.rsh(node, self.templates['CibDelXpath'] % rsc)



"""Restart all nodes in order."""

__all__ = ["RestartOnebyOne"]
__copyright__ = "Copyright 2000-2024 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

from pacemaker._cts.tests.ctstest import CTSTest
from pacemaker._cts.tests.restarttest import RestartTest
from pacemaker._cts.tests.simulstartlite import SimulStartLite

# Disable various pylint warnings that occur in so many places throughout this
# file it's easiest to just take care of them globally.  This does introduce the
# possibility that we'll miss some other cause of the same warning, but we'll
# just have to be careful.

# pylint doesn't understand that self._env is subscriptable.
# pylint: disable=unsubscriptable-object


class RestartOnebyOne(CTSTest):
    """Restart all nodes in order."""

    def __init__(self, cm):
        """
        Create a new RestartOnebyOne instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        CTSTest.__init__(self, cm)

        self.name = "RestartOnebyOne"

        self._restart = None
        self._startall = SimulStartLite(cm)

    def __call__(self, dummy):
        """Perform the test."""
        self.incr("calls")

        ret = self._startall(None)
        if not ret:
            return self.failure("Setup failed")

        did_fail = []
        self.set_timer()
        self._restart = RestartTest(self._cm)

        for node in self._env["nodes"]:
            if not self._restart(node):
                did_fail.append(node)

        if did_fail:
            return self.failure("Could not restart %d nodes: %r" % (len(did_fail), did_fail))

        return self.success()



"""Fence a running node and wait for it to restart."""

__all__ = ["StonithdTest"]
__copyright__ = "Copyright 2000-2024 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

from pacemaker.exitstatus import ExitStatus
from pacemaker._cts.tests.ctstest import CTSTest
from pacemaker._cts.tests.simulstartlite import SimulStartLite
from pacemaker._cts.timer import Timer

# Disable various pylint warnings that occur in so many places throughout this
# file it's easiest to just take care of them globally.  This does introduce the
# possibility that we'll miss some other cause of the same warning, but we'll
# just have to be careful.

# pylint doesn't understand that self._rsh is callable.
# pylint: disable=not-callable
# pylint doesn't understand that self._env is subscriptable.
# pylint: disable=unsubscriptable-object


class StonithdTest(CTSTest):
    """Fence a running node and wait for it to restart."""

    def __init__(self, cm):
        """
        Create a new StonithdTest instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        CTSTest.__init__(self, cm)
        self.benchmark = True
        self.name = "Stonithd"

        self._startall = SimulStartLite(cm)

    def __call__(self, node):
        """Perform this test."""
        self.incr("calls")
        if len(self._env["nodes"]) < 2:
            return self.skipped()

        ret = self._startall(None)
        if not ret:
            return self.failure("Setup failed")

        watchpats = [
            self.templates["Pat:Fencing_ok"] % node,
            self.templates["Pat:NodeFenced"] % node,
        ]

        if not self._env["at-boot"]:
            self.debug("Expecting %s to stay down" % node)
            self._cm.expected_status[node] = "down"
        else:
            self.debug("Expecting %s to come up again %d" % (node, self._env["at-boot"]))
            watchpats.extend([
                "%s.* S_STARTING -> S_PENDING" % node,
                "%s.* S_PENDING -> S_NOT_DC" % node,
            ])

        watch = self.create_watch(watchpats, 30 + self._env["DeadTime"] + self._env["StableTime"] + self._env["StartTime"])
        watch.set_watch()

        origin = self._env.random_gen.choice(self._env["nodes"])

        (rc, _) = self._rsh(origin, "stonith_admin --reboot %s -VVVVVV" % node)

        if rc == ExitStatus.TIMEOUT:
            # Look for the patterns, usually this means the required
            # device was running on the node to be fenced - or that
            # the required devices were in the process of being loaded
            # and/or moved
            #
            # Effectively the node committed suicide so there will be
            # no confirmation, but pacemaker should be watching and
            # fence the node again

            self._logger.log("Fencing command on %s to fence %s timed out" % (origin, node))

        elif origin != node and rc != 0:
            self.debug("Waiting for the cluster to recover")
            self._cm.cluster_stable()

            self.debug("Waiting for fenced node to come back up")
            self._cm.ns.wait_for_all_nodes(self._env["nodes"], 600)

            self._logger.log("Fencing command on %s failed to fence %s (rc=%d)" % (origin, node, rc))

        elif origin == node and rc != 255:
            # 255 == broken pipe, ie. the node was fenced as expected
            self._logger.log("Locally originated fencing returned %d" % rc)

        with Timer(self._logger, self.name, "fence"):
            matched = watch.look_for_all()

        self.set_timer("reform")
        if watch.unmatched:
            self._logger.log("Patterns not found: %r" % watch.unmatched)

        self.debug("Waiting for the cluster to recover")
        self._cm.cluster_stable()

        self.debug("Waiting for fenced node to come back up")
        self._cm.ns.wait_for_all_nodes(self._env["nodes"], 600)

        self.debug("Waiting for the cluster to re-stabilize with all nodes")
        is_stable = self._cm.cluster_stable(self._env["StartTime"])

        if not matched:
            return self.failure("Didn't find all expected patterns")

        if not is_stable:
            return self.failure("Cluster did not become stable")

        self.log_timer("reform")
        return self.success()

    @property
    def errors_to_ignore(self):
        """Return a list of errors which should be ignored."""
        return [
            self.templates["Pat:Fencing_start"] % ".*",
            self.templates["Pat:Fencing_ok"] % ".*",
            self.templates["Pat:Fencing_active"],
            r"error.*: Operation 'reboot' targeting .* by .* for stonith_admin.*: Timer expired"
        ]

    def is_applicable(self):
        """Return True if this test is applicable in the current test configuration."""
        if not CTSTest.is_applicable(self):
            return False

        # pylint gets confused because of EnvFactory here.
        # pylint: disable=unsupported-membership-test
        if "DoFencing" in self._env:
            return self._env["DoFencing"]

        return True



"""Simultaneously start stopped nodes."""

__all__ = ["SimulStartLite"]
__copyright__ = "Copyright 2000-2024 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

from pacemaker._cts.tests.ctstest import CTSTest

# Disable various pylint warnings that occur in so many places throughout this
# file it's easiest to just take care of them globally.  This does introduce the
# possibility that we'll miss some other cause of the same warning, but we'll
# just have to be careful.

# pylint doesn't understand that self._env is subscriptable.
# pylint: disable=unsubscriptable-object


class SimulStartLite(CTSTest):
    """
    A pseudo-test that sets up conditions before running some other test.

    This class starts any stopped nodes more or less simultaneously.  Other test
    classes should not use this one as a superclass.
    """

    def __init__(self, cm):
        """
        Create a new SimulStartLite instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        CTSTest.__init__(self, cm)
        self.name = "SimulStartLite"

    def __call__(self, dummy):
        """Return whether starting all stopped nodes more or less simultaneously succeeds."""
        self.incr("calls")
        self.debug("Setup: %s" % self.name)

        # We ignore the "node" parameter...
        node_list = []
        for node in self._env["nodes"]:
            if self._cm.expected_status[node] == "down":
                self.incr("WasStopped")
                node_list.append(node)

        self.set_timer()
        while len(node_list) > 0:
            # Repeat until all nodes come up
            uppat = self.templates["Pat:NonDC_started"]
            if self._cm.upcount() == 0:
                uppat = self.templates["Pat:Local_started"]

            watchpats = [
                self.templates["Pat:DC_IDLE"]
            ]
            for node in node_list:
                watchpats.extend([uppat % node,
                                  self.templates["Pat:InfraUp"] % node,
                                  self.templates["Pat:PacemakerUp"] % node])

            #   Start all the nodes - at about the same time...
            watch = self.create_watch(watchpats, self._env["DeadTime"] + 10)
            watch.set_watch()

            stonith = self._cm.prepare_fencing_watcher()

            for node in node_list:
                self._cm.start_cm_async(node)

            watch.look_for_all()

            node_list = self._cm.fencing_cleanup(self.name, stonith)

            if node_list is None:
                return self.failure("Cluster did not stabilize")

            # Remove node_list messages from watch.unmatched
            for node in node_list:
                self._logger.debug("Dealing with stonith operations for %s" % node_list)
                if watch.unmatched:
                    try:
                        watch.unmatched.remove(uppat % node)
                    except ValueError:
                        self.debug("Already matched: %s" % (uppat % node))

                    try:
                        watch.unmatched.remove(self.templates["Pat:InfraUp"] % node)
                    except ValueError:
                        self.debug("Already matched: %s" % (self.templates["Pat:InfraUp"] % node))

                    try:
                        watch.unmatched.remove(self.templates["Pat:PacemakerUp"] % node)
                    except ValueError:
                        self.debug("Already matched: %s" % (self.templates["Pat:PacemakerUp"] % node))

            if watch.unmatched:
                for regex in watch.unmatched:
                    self._logger.log("Warn: Startup pattern not found: %s" % regex)

            if not self._cm.cluster_stable():
                return self.failure("Cluster did not stabilize")

        did_fail = False
        unstable = []
        for node in self._env["nodes"]:
            if not self._cm.stat_cm(node):
                did_fail = True
                unstable.append(node)

        if did_fail:
            return self.failure("Unstarted nodes exist: %s" % unstable)

        unstable = []
        for node in self._env["nodes"]:
            if not self._cm.node_stable(node):
                did_fail = True
                unstable.append(node)

        if did_fail:
            return self.failure("Unstable cluster nodes exist: %s" % unstable)

        return self.success()

    def is_applicable(self):
        """Return True if this test is applicable in the current test configuration."""
        return False



"""CIB generator for Pacemaker's Cluster Test Suite (CTS)."""

__all__ = ["ConfigFactory"]
__copyright__ = "Copyright 2008-2024 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

import warnings
import tempfile

from pacemaker.buildoptions import BuildOptions
from pacemaker._cts.cibxml import Alerts, Clone, Expression, FencingTopology, Group, Nodes, OpDefaults, Option, Resource, Rule
from pacemaker._cts.network import next_ip


class CIB:
    """A class for generating, representing, and installing a CIB file onto cluster nodes."""

    def __init__(self, cm, version, factory, tmpfile=None):
        """
        Create a new CIB instance.

        Arguments:
        cm      -- A ClusterManager instance
        version -- The schema syntax version
        factory -- A ConfigFactory instance
        tmpfile -- Where to store the CIB, or None to use a new tempfile
        """
        # pylint: disable=invalid-name
        self._cib = None
        self._cm = cm
        self._counter = 1
        self._factory = factory
        self._num_nodes = 0

        self.version = version

        if not tmpfile:
            warnings.filterwarnings("ignore")

            # pylint: disable=consider-using-with
            f = tempfile.NamedTemporaryFile(delete=True)
            f.close()
            tmpfile = f.name

            warnings.resetwarnings()

        self._factory.tmpfile = tmpfile

    def _show(self):
        """Query a cluster node for its generated CIB; log and return the result."""
        output = ""
        (_, result) = self._factory.rsh(self._factory.target, "HOME=/root CIB_file=%s cibadmin -Ql" % self._factory.tmpfile, verbose=1)

        for line in result:
            output += line
            self._factory.debug("Generated Config: %s" % line)

        return output

    def new_ip(self, name=None):
        """Generate an IP resource for the next available IP address, optionally specifying the resource's name."""
        if self._cm.env["IPagent"] == "IPaddr2":
            ip = next_ip(self._cm.env["IPBase"])
            if not name:
                if ":" in ip:
                    (_, _, suffix) = ip.rpartition(":")
                    name = "r%s" % suffix
                else:
                    name = "r%s" % ip

            r = Resource(self._factory, name, self._cm.env["IPagent"], "ocf")
            r["ip"] = ip

            if ":" in ip:
                r["cidr_netmask"] = "64"
                r["nic"] = "eth0"
            else:
                r["cidr_netmask"] = "32"

        else:
            if not name:
                name = "r%s%d" % (self._cm.env["IPagent"], self._counter)
                self._counter += 1

            r = Resource(self._factory, name, self._cm.env["IPagent"], "ocf")

        r.add_op("monitor", "5s")
        return r

    def get_node_id(self, node_name):
        """Check the cluster configuration for the node ID for the given node_name."""
        # We can't account for every possible configuration,
        # so we only return a node ID if:
        # * The node is specified in /etc/corosync/corosync.conf
        #   with "ring0_addr:" equal to node_name and "nodeid:"
        #   explicitly specified.
        # In all other cases, we return 0.
        node_id = 0

        # awkward command: use } as record separator
        # so each corosync.conf "object" is one record;
        # match the "node {" record that has "ring0_addr: node_name";
        # then print the substring of that record after "nodeid:"
        awk = r"""awk -v RS="}" """ \
              r"""'/^(\s*nodelist\s*{)?\s*node\s*{.*(ring0_addr|name):\s*%s(\s+|$)/""" \
              r"""{gsub(/.*nodeid:\s*/,"");gsub(/\s+.*$/,"");print}' %s""" \
              % (node_name, BuildOptions.COROSYNC_CONFIG_FILE)

        (rc, output) = self._factory.rsh(self._factory.target, awk, verbose=1)

        if rc == 0 and len(output) == 1:
            try:
                node_id = int(output[0])
            except ValueError:
                node_id = 0

        return node_id

    def install(self, target):
        """Generate a CIB file and install it to the given cluster node."""
        old = self._factory.tmpfile

        # Force a rebuild
        self._cib = None

        self._factory.tmpfile = "%s/cib.xml" % BuildOptions.CIB_DIR
        self.contents(target)
        self._factory.rsh(self._factory.target, "chown %s %s" % (BuildOptions.DAEMON_USER, self._factory.tmpfile))

        self._factory.tmpfile = old

    def contents(self, target):
        """Generate a complete CIB file."""
        # fencing resource
        if self._cib:
            return self._cib

        if target:
            self._factory.target = target

        self._factory.rsh(self._factory.target, "HOME=/root cibadmin --empty %s > %s" % (self.version, self._factory.tmpfile))
        self._num_nodes = len(self._cm.env["nodes"])

        no_quorum = "stop"
        if self._num_nodes < 3:
            no_quorum = "ignore"
            self._factory.log("Cluster only has %d nodes, configuring: no-quorum-policy=ignore" % self._num_nodes)

        # We don't need a nodes section unless we add attributes
        stn = None

        # Fencing resource
        # Define first so that the shell doesn't reject every update
        if self._cm.env["DoFencing"]:

            # Define the "real" fencing device
            st = Resource(self._factory, "Fencing", self._cm.env["stonith-type"], "stonith")

            # Set a threshold for unreliable stonith devices such as the vmware one
            st.add_meta("migration-threshold", "5")
            st.add_op("monitor", "120s", timeout="120s")
            st.add_op("stop", "0", timeout="60s")
            st.add_op("start", "0", timeout="60s")

            # For remote node tests, a cluster node is stopped and brought back up
            # as a remote node with the name "remote-OLDNAME". To allow fencing
            # devices to fence these nodes, create a list of all possible node names.
            all_node_names = [prefix + n for n in self._cm.env["nodes"] for prefix in ('', 'remote-')]

            # Add all parameters specified by user
            entries = self._cm.env["stonith-params"].split(',')
            for entry in entries:
                try:
                    (name, value) = entry.split('=', 1)
                except ValueError:
                    print("Warning: skipping invalid fencing parameter: %s" % entry)
                    continue

                # Allow user to specify "all" as the node list, and expand it here
                if name in ["hostlist", "pcmk_host_list"] and value == "all":
                    value = ' '.join(all_node_names)

                st[name] = value

            st.commit()

            # Test advanced fencing logic
            stf_nodes = []
            stt_nodes = []
            attr_nodes = {}

            # Create the levels
            stl = FencingTopology(self._factory)
            for node in self._cm.env["nodes"]:
                # Remote node tests will rename the node
                remote_node = "remote-%s" % node

                # Randomly assign node to a fencing method
                ftype = self._cm.env.random_gen.choice(["levels-and", "levels-or ", "broadcast "])

                # For levels-and, randomly choose targeting by node name or attribute
                by = ""

                if ftype == "levels-and":
                    node_id = self.get_node_id(node)

                    if node_id == 0 or self._cm.env.random_gen.choice([True, False]):
                        by = " (by name)"
                    else:
                        attr_nodes[node] = node_id
                        by = " (by attribute)"

                self._cm.log(" - Using %s fencing for node: %s%s" % (ftype, node, by))

                if ftype == "levels-and":
                    # If targeting by name, add a topology level for this node
                    if node not in attr_nodes:
                        stl.level(1, node, "FencingPass,Fencing")

                    # Always target remote nodes by name, otherwise we would need to add
                    # an attribute to the remote node only during remote tests (we don't
                    # want nonexistent remote nodes showing up in the non-remote tests).
                    # That complexity is not worth the effort.
                    stl.level(1, remote_node, "FencingPass,Fencing")

                    # Add the node (and its remote equivalent) to the list of levels-and nodes.
                    stt_nodes.extend([node, remote_node])

                elif ftype == "levels-or ":
                    for n in [node, remote_node]:
                        stl.level(1, n, "FencingFail")
                        stl.level(2, n, "Fencing")

                    stf_nodes.extend([node, remote_node])

            # If any levels-and nodes were targeted by attribute,
            # create the attributes and a level for the attribute.
            if attr_nodes:
                stn = Nodes(self._factory)

                for (node_name, node_id) in attr_nodes.items():
                    stn.add_node(node_name, node_id, {"cts-fencing": "levels-and"})

                stl.level(1, None, "FencingPass,Fencing", "cts-fencing", "levels-and")

            # Create a Dummy agent that always passes for levels-and
            if stt_nodes:
                stt = Resource(self._factory, "FencingPass", "fence_dummy", "stonith")
                stt["pcmk_host_list"] = " ".join(stt_nodes)
                # Wait this many seconds before doing anything, handy for letting disks get flushed too
                stt["random_sleep_range"] = "30"
                stt["mode"] = "pass"
                stt.commit()

            # Create a Dummy agent that always fails for levels-or
            if stf_nodes:
                stf = Resource(self._factory, "FencingFail", "fence_dummy", "stonith")
                stf["pcmk_host_list"] = " ".join(stf_nodes)
                # Wait this many seconds before doing anything, handy for letting disks get flushed too
                stf["random_sleep_range"] = "30"
                stf["mode"] = "fail"
                stf.commit()

            # Now commit the levels themselves
            stl.commit()

        o = Option(self._factory)
        o["stonith-enabled"] = self._cm.env["DoFencing"]
        o["start-failure-is-fatal"] = "false"
        o["pe-input-series-max"] = "5000"
        o["shutdown-escalation"] = "5min"
        o["batch-limit"] = "10"
        o["dc-deadtime"] = "5s"
        o["no-quorum-policy"] = no_quorum

        o.commit()

        o = OpDefaults(self._factory)
        o["timeout"] = "90s"
        o.commit()

        # Commit the nodes section if we defined one
        if stn is not None:
            stn.commit()

        # Add an alerts section if possible
        if self._factory.rsh.exists_on_all(self._cm.env["notification-agent"], self._cm.env["nodes"]):
            alerts = Alerts(self._factory)
            alerts.add_alert(self._cm.env["notification-agent"],
                             self._cm.env["notification-recipient"])
            alerts.commit()

        # Add resources?
        if self._cm.env["CIBResource"]:
            self.add_resources()

        # generate cib
        self._cib = self._show()

        if self._factory.tmpfile != "%s/cib.xml" % BuildOptions.CIB_DIR:
            self._factory.rsh(self._factory.target, "rm -f %s" % self._factory.tmpfile)

        return self._cib

    def add_resources(self):
        """Add various resources and their constraints to the CIB."""
        # Per-node resources
        for node in self._cm.env["nodes"]:
            name = "rsc_%s" % node
            r = self.new_ip(name)
            r.prefer(node, "100")
            r.commit()

        # Migrator
        # Make this slightly sticky (since we have no other location constraints) to avoid relocation during Reattach
        m = Resource(self._factory, "migrator", "Dummy", "ocf", "pacemaker")
        m["passwd"] = "whatever"
        m.add_meta("resource-stickiness", "1")
        m.add_meta("allow-migrate", "1")
        m.add_op("monitor", "P10S")
        m.commit()

        # Ping the test exerciser
        p = Resource(self._factory, "ping-1", "ping", "ocf", "pacemaker")
        p.add_op("monitor", "60s")
        p["host_list"] = self._cm.env["cts-exerciser"]
        p["name"] = "connected"
        p["debug"] = "true"

        c = Clone(self._factory, "Connectivity", p)
        c["globally-unique"] = "false"
        c.commit()

        # promotable clone resource
        s = Resource(self._factory, "stateful-1", "Stateful", "ocf", "pacemaker")
        s.add_op("monitor", "15s", timeout="60s")
        s.add_op("monitor", "16s", timeout="60s", role="Promoted")
        ms = Clone(self._factory, "promotable-1", s)
        ms["promotable"] = "true"
        ms["clone-max"] = self._num_nodes
        ms["clone-node-max"] = 1
        ms["promoted-max"] = 1
        ms["promoted-node-max"] = 1

        # Require connectivity to run the promotable clone
        r = Rule(self._factory, "connected", "-INFINITY", op="or")
        r.add_child(Expression(self._factory, "m1-connected-1", "connected", "lt", "1"))
        r.add_child(Expression(self._factory, "m1-connected-2", "connected", "not_defined", None))
        ms.prefer("connected", rule=r)

        ms.commit()

        # Group Resource
        g = Group(self._factory, "group-1")
        g.add_child(self.new_ip())

        if self._cm.env["have_systemd"]:
            sysd = Resource(self._factory, "petulant", "pacemaker-cts-dummyd@10", "service")
            sysd.add_op("monitor", "P10S")
            g.add_child(sysd)
        else:
            g.add_child(self.new_ip())

        g.add_child(self.new_ip())

        # Make group depend on the promotable clone
        g.after("promotable-1", first="promote", then="start")
        g.colocate("promotable-1", "INFINITY", withrole="Promoted")

        g.commit()

        # LSB resource
        lsb = Resource(self._factory, "lsb-dummy", "LSBDummy", "lsb")
        lsb.add_op("monitor", "5s")

        # LSB with group
        lsb.after("group-1")
        lsb.colocate("group-1")

        lsb.commit()


class ConfigFactory:
    """Singleton to generate a CIB file for the environment's schema version."""

    def __init__(self, cm):
        """
        Create a new ConfigFactory instance.

        Arguments:
        cm      -- A ClusterManager instance
        """
        # pylint: disable=invalid-name
        self._cm = cm
        self.rsh = self._cm.rsh
        if not self._cm.env["ListTests"]:
            self.target = self._cm.env["nodes"][0]
        self.tmpfile = None

    def log(self, args):
        """Log a message."""
        self._cm.log("cib: %s" % args)

    def debug(self, args):
        """Log a debug message."""
        self._cm.debug("cib: %s" % args)

    def create_config(self, name="pacemaker-%s" % BuildOptions.CIB_SCHEMA_VERSION):
        """Return a CIB object for the given schema version."""
        return CIB(self._cm, name, self)



"""User input related utilities for CTS."""

__all__ = ["should_continue"]
__copyright__ = "Copyright 2000-2024 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"


def should_continue(env):
    """On failure, prompt the user to see if we should continue."""
    if env["continue"]:
        return True

    try:
        answer = input("Continue? [yN]")
    except EOFError:
        answer = "n"

    return answer in ["y", "Y"]
